import unittest
import mock
import logging
import json
import sys
from tests.utils import compare_json_return_errormessage as error_message
import tests.utils

import esgfpid.assistant.unpublish

LOGGER = logging.getLogger(__name__)
LOGGER.addHandler(logging.NullHandler())

# Test resources:
from resources.TESTVALUES import *
import resources.TESTVALUES as TESTHELPERS

class UnpublicationTestCase(unittest.TestCase):

    def tearDown(self):
        LOGGER.info('#############################')

    def setUp(self):
        LOGGER.info('######## Next test (%s) ##########', __name__)

    def test_unpublish_one_version_by_handle(self):

        # Preparations
        testcoupler = TESTHELPERS.get_coupler(solr_switched_off=True)
        TESTHELPERS.patch_with_rabbit_mock(testcoupler)
        args = TESTHELPERS.get_args_for_unpublish_one()
        args['drs_id'] = None # not passed
        assistant = esgfpid.assistant.unpublish.AssistantOneVersion(coupler=testcoupler, **args)
  
        # Run code to be tested:
        assistant.unpublish_one_dataset_version(
            dataset_handle=DATASETHANDLE_HDL # is redundant, but will be checked.
        )

        # Check result:
        expected_rabbit_task = TESTHELPERS.get_rabbit_message_unpub_one_by_handle()
        received_rabbit_task = TESTHELPERS.get_received_message_from_rabbitmock(testcoupler)
        same = utils.is_json_same(expected_rabbit_task, received_rabbit_task)
        self.assertTrue(same, error_message(expected_rabbit_task, received_rabbit_task))

    def test_unpublish_one_version_by_version_number(self):

        # Preparations
        testcoupler = TESTHELPERS.get_coupler(solr_switched_off=True)
        TESTHELPERS.patch_with_rabbit_mock(testcoupler)
        args = TESTHELPERS.get_args_for_unpublish_one()
        assistant = esgfpid.assistant.unpublish.AssistantOneVersion(coupler=testcoupler, **args)
  
        # Run code to be tested:
        assistant.unpublish_one_dataset_version(
            version_number=DS_VERSION
        )

        # Check result:
        expected_rabbit_task = TESTHELPERS.get_rabbit_message_unpub_one()
        received_rabbit_task = TESTHELPERS.get_received_message_from_rabbitmock(testcoupler)
        same = utils.is_json_same(expected_rabbit_task, received_rabbit_task)
        self.assertTrue(same, error_message(expected_rabbit_task, received_rabbit_task))

    def test_unpublish_one_version_by_version_number_and_handle(self):
 
        # Preparations
        testcoupler = TESTHELPERS.get_coupler(solr_switched_off=True)
        TESTHELPERS.patch_with_rabbit_mock(testcoupler)
        args = TESTHELPERS.get_args_for_unpublish_one()
        assistant = esgfpid.assistant.unpublish.AssistantOneVersion(coupler=testcoupler, **args)

        # Run code to be tested:
        assistant.unpublish_one_dataset_version(
            version_number=DS_VERSION,
            dataset_handle=DATASETHANDLE_HDL # is redundant, but will be checked.
        )

        # Check result:
        expected_rabbit_task = TESTHELPERS.get_rabbit_message_unpub_one()
        received_rabbit_task = TESTHELPERS.get_received_message_from_rabbitmock(testcoupler)
        same = utils.is_json_same(expected_rabbit_task, received_rabbit_task)
        self.assertTrue(same, error_message(expected_rabbit_task, received_rabbit_task))

    def test_unpublish_one_version_wrong_handle(self):

        # Test variables
        version_number = '999888'
        wrong_handle = PREFIX_NO_HDL+'/miauz'
  
        # Preparations
        testcoupler = TESTHELPERS.get_coupler(solr_switched_off=True)
        TESTHELPERS.patch_with_rabbit_mock(testcoupler)
        args = TESTHELPERS.get_args_for_unpublish_one()
        assistant = esgfpid.assistant.unpublish.AssistantOneVersion(coupler=testcoupler, **args)

        # Run code to be tested and check exception:
        with self.assertRaises(ValueError):
            assistant.unpublish_one_dataset_version(
                version_number=version_number,
                dataset_handle=wrong_handle)

    def test_unpublish_one_version_version_is_none(self):
        
        # Test variables
        version_number = None

        # Preparations
        testcoupler = TESTHELPERS.get_coupler(solr_switched_off=True)
        TESTHELPERS.patch_with_rabbit_mock(testcoupler)
        args = TESTHELPERS.get_args_for_unpublish_one()
        assistant = esgfpid.assistant.unpublish.AssistantOneVersion(coupler=testcoupler, **args)

        # Run code to be tested and check exception:
        with self.assertRaises(esgfpid.exceptions.ArgumentError):
            assistant.unpublish_one_dataset_version(version_number=version_number)

    def test_unpublish_one_version_no_version_given(self):
        
        # Preparations
        testcoupler = TESTHELPERS.get_coupler(solr_switched_off=True)
        TESTHELPERS.patch_with_rabbit_mock(testcoupler)
        args = TESTHELPERS.get_args_for_unpublish_one()
        assistant = esgfpid.assistant.unpublish.AssistantOneVersion(coupler=testcoupler, **args)

        # Run code to be tested and check exception:
        with self.assertRaises(esgfpid.exceptions.ArgumentError):
            assistant.unpublish_one_dataset_version()

    def test_unpublish_all_versions_nosolr_consumer_must_find_versions_ok(self):        

        # Preparations:
        # Preparations
        testcoupler = TESTHELPERS.get_coupler(solr_switched_off=True)
        TESTHELPERS.patch_with_rabbit_mock(testcoupler)
        args = TESTHELPERS.get_args_for_unpublish_all()
        args['consumer_solr_url']=SOLR_URL_CONSUMER
        assistant = esgfpid.assistant.unpublish.AssistantAllVersions(coupler=testcoupler, **args)

        # Run code to be tested:
        assistant.unpublish_all_dataset_versions()

        # Check result:
        expected_rabbit_task = TESTHELPERS.get_rabbit_message_unpub_all()
        expected_rabbit_task["consumer_solr_url"] = SOLR_URL_CONSUMER
        received_rabbit_task = TESTHELPERS.get_received_message_from_rabbitmock(testcoupler)
        same = utils.is_json_same(expected_rabbit_task, received_rabbit_task)
        self.assertTrue(same, error_message(expected_rabbit_task, received_rabbit_task))

    def test_unpublish_all_versions_solr_off_consumer_must_find_versions_ok(self):        

        # Preparations:
        testcoupler = TESTHELPERS.get_coupler(solr_url=None, solr_switched_off=True)
        TESTHELPERS.patch_with_rabbit_mock(testcoupler)
        args = TESTHELPERS.get_args_for_unpublish_all()
        args["consumer_solr_url"]=SOLR_URL_CONSUMER
        assistant = esgfpid.assistant.unpublish.AssistantAllVersions(coupler=testcoupler, **args)

        # Run code to be tested:
        assistant.unpublish_all_dataset_versions()

        # Check result:
        expected_rabbit_task = TESTHELPERS.get_rabbit_message_unpub_all()
        expected_rabbit_task["consumer_solr_url"] = SOLR_URL_CONSUMER
        received_rabbit_task = TESTHELPERS.get_received_message_from_rabbitmock(testcoupler)
        same = utils.is_json_same(expected_rabbit_task, received_rabbit_task)
        self.assertTrue(same, error_message(expected_rabbit_task, received_rabbit_task))

    def test_unpublish_all_versions_by_handles_ok(self):

        # Test variables
        list_of_dataset_handles = [PREFIX_NO_HDL+'/bla', PREFIX_NO_HDL+'/blub']

        # Set solr mock to return handles:
        testcoupler = TESTHELPERS.get_coupler()
        TESTHELPERS.patch_solr_returns_list_of_datasets_and_versions(testcoupler, list_of_dataset_handles, None)
        TESTHELPERS.patch_with_rabbit_mock(testcoupler)
        args = TESTHELPERS.get_args_for_unpublish_all()
        assistant = esgfpid.assistant.unpublish.AssistantAllVersions(coupler=testcoupler, **args)

        # Run code to be tested:
        assistant.unpublish_all_dataset_versions()

        # Check result:
        received_rabbit_task1 = TESTHELPERS.get_received_message_from_rabbitmock(testcoupler, 0)
        received_rabbit_task2 = TESTHELPERS.get_received_message_from_rabbitmock(testcoupler, 1)
        expected_rabbit_task1 = TESTHELPERS.get_rabbit_message_unpub_one()
        expected_rabbit_task1["handle"]=list_of_dataset_handles[0]
        del expected_rabbit_task1["version_number"]
        expected_rabbit_task2 = TESTHELPERS.get_rabbit_message_unpub_one()
        expected_rabbit_task2["handle"]=list_of_dataset_handles[1]
        del expected_rabbit_task2["version_number"]
        same1 = utils.is_json_same(expected_rabbit_task1, received_rabbit_task1)
        same2 = utils.is_json_same(expected_rabbit_task2, received_rabbit_task2)
        self.assertTrue(same1, error_message(expected_rabbit_task1, received_rabbit_task1))
        self.assertTrue(same2, error_message(expected_rabbit_task2, received_rabbit_task2))

    def test_unpublish_all_versions_by_version_numbers_ok(self):
        
        # Test variables
        list_of_version_numbers = [DS_VERSION, DS_VERSION2]

        # Set solr mock to return handles:
        testcoupler = TESTHELPERS.get_coupler()
        TESTHELPERS.patch_solr_returns_list_of_datasets_and_versions(testcoupler, None, list_of_version_numbers)
        TESTHELPERS.patch_with_rabbit_mock(testcoupler)
        args = TESTHELPERS.get_args_for_unpublish_all()
        assistant = esgfpid.assistant.unpublish.AssistantAllVersions(coupler=testcoupler, **args)
        #self.coupler._Coupler__solr_sender.datasethandles_or_versionnumbers_of_allversions['version_numbers'] = list_of_version_numbers

        # Run code to be tested:
        assistant.unpublish_all_dataset_versions()

        # Check result:
        expected_rabbit_task1 = {
            "operation": "unpublish_one_version",
            "aggregation_level": "dataset",
            "message_timestamp": "anydate",
            "data_node": DATA_NODE,
            "handle": DATASETHANDLE_HDL,
            "ROUTING_KEY": PREFIX_FOR_ROUTINGKEY+'.HASH.fresh.unpubli-onevers',
            "drs_id": DRS_ID,
            "version_number": DS_VERSION
        }
        expected_rabbit_task2 = {
            "operation": "unpublish_one_version",
            "aggregation_level": "dataset",
            "message_timestamp": "anydate",
            "data_node": DATA_NODE,
            "handle": DATASETHANDLE_HDL2,
            "ROUTING_KEY": PREFIX_FOR_ROUTINGKEY+'.HASH.fresh.unpubli-onevers',
            "drs_id": DRS_ID,
            "version_number": DS_VERSION2
        }
        received_rabbit_task1 = TESTHELPERS.get_received_message_from_rabbitmock(testcoupler, 0)
        received_rabbit_task2 = TESTHELPERS.get_received_message_from_rabbitmock(testcoupler, 1)
        same1 = utils.is_json_same(expected_rabbit_task1, received_rabbit_task1)
        same2 = utils.is_json_same(expected_rabbit_task2, received_rabbit_task2)
        self.assertTrue(same1, error_message(expected_rabbit_task1, received_rabbit_task1))
        self.assertTrue(same2, error_message(expected_rabbit_task2, received_rabbit_task2))


import unittest
import mock
import logging
import pika.exceptions as pikaexceptions
import tests.utils.captureconsoleoutput
import tests.resources.pikamock
import tests.resources.error_messages
import esgfpid.utils
import esgfpid.check
import pika

# Logging
LOGGER = logging.getLogger(__name__)
LOGGER.addHandler(logging.NullHandler())

# Test resources:
from resources.TESTVALUES import *
import resources.TESTVALUES as TESTHELPERS

import globalvar
if globalvar.QUICK_ONLY:
    print('Skipping slow tests in module "%s".' % __name__)


class CheckTestCase(unittest.TestCase):

    slow_message = '\nRunning a slow test (avoid by using -ls flag).'

    def setUp(self):
        LOGGER.info('######## Next test (%s) ##########', __name__)

    def tearDown(self):
        LOGGER.info('#############################')

    #
    # Init
    #

    def test_init_rabbitchecker_no_preferred(self):
        ''' Test if the constructor of the RabbitChecker works fine (with foo bar values).'''

        # Test variables:
        rabbit1 = dict(
            user = 'johndoe',
            password = 'abc123yx',
            url = 'this.is.my.only.host')
        testconnector = TESTHELPERS.get_connector(messaging_service_credentials=[rabbit1])

        # Run code to be tested:
        testobject = esgfpid.check.RabbitChecker(
            connector = testconnector,
            print_to_console = True,
            print_success_to_console = True
        )

        # Check result:
        self.assertIsInstance(testobject, esgfpid.check.RabbitChecker,
            'Init with placeholder params did not work.')

    def test_init_rabbitchecker_missing_args(self):
        ''' Test if the constructor of the RabbitChecker complains if an argument is missing.'''

        # Run code to be tested and check exception:
        with self.assertRaises(esgfpid.exceptions.ArgumentError):
            testobject = esgfpid.check.RabbitChecker(
                print_to_console = True,
                print_success_to_console = True
            )

    #
    # Positive check
    #

    @mock.patch('esgfpid.check.RabbitChecker._RabbitChecker__pika_blocking_connection')
    def test_run_check_ok(self, connection_patch):
        ''' This checks if the entire function works fine in case
        the rabbit is reacked (connection and channel ok), and if it prints 
        the correct message.
        '''

        # Define the replacement for the patched method:
        mock_response = tests.resources.pikamock.MockPikaBlockingConnection(mock.MagicMock())
        connection_patch.return_value = mock_response

        # Test variables:
        rabbit1 = dict(
            user = 'johndoe',
            password = 'abc123yx',
            url = 'this.is.my.only.host')
        testconnector = TESTHELPERS.get_connector(messaging_service_credentials=[rabbit1])


        # Run code to be tested and capture stout:
        with tests.utils.captureconsoleoutput.captured_output() as (out, err):
            returned_msg = esgfpid.check.check_pid_queue_availability(
                connector = testconnector,
                print_to_console = True,
                print_success_to_console = True
            )

        # Expected message:
        expected_message = tests.resources.error_messages.expected_message_ok1
        expected_return = tests.resources.error_messages.expected_message_ok2

        # Check result:
        output = out.getvalue().strip()
        self.assertEqual(expected_message, output,
            'Wrong stdout message.\n\nWe expected:\n\n%s\n\nWe got:\n\n%s\n' % (expected_message, output))
        self.assertEqual(expected_return, returned_msg,
            'Wrong return message.\n\nWe expected:\n\n%s\n\nWe got:\n\n%s\n' % (expected_return, returned_msg))

    @mock.patch('esgfpid.check.RabbitChecker._RabbitChecker__pika_blocking_connection')
    def test_run_check_with_msg_ok(self, connection_patch):
        ''' This checks if the entire function works fine in case
        the rabbit is reacked (connection and channel ok), and if it prints 
        the correct message.
        '''

        # Define the replacement for the patched method:
        mock_response = tests.resources.pikamock.MockPikaBlockingConnection(mock.MagicMock())
        connection_patch.return_value = mock_response

        # Test variables:
        rabbit1 = dict(
            user = 'johndoe',
            password = 'abc123yx',
            url = 'this.is.my.only.host')
        testconnector = TESTHELPERS.get_connector(messaging_service_credentials=[rabbit1])


        # Run code to be tested and capture stout:
        with tests.utils.captureconsoleoutput.captured_output() as (out, err):
            returned_msg = esgfpid.check.check_pid_queue_availability(
                connector = testconnector,
                print_to_console = True,
                print_success_to_console = True,
                send_message = True
            )

        # Expected message:
        expected_message = tests.resources.error_messages.expected_message_plus_testmsg_ok1
        expected_return = tests.resources.error_messages.expected_message_plus_testmsg_ok2

        # Check result:
        output = out.getvalue().strip()
        self.assertEqual(expected_message, output,
            'Wrong stdout message.\n\nWe expected:\n\n%s\n\nWe got:\n\n%s\n' % (expected_message, output))
        self.assertEqual(expected_return, returned_msg,
            'Wrong return message.\n\nWe expected:\n\n%s\n\nWe got:\n\n%s\n' % (expected_return, returned_msg))

    #
    # Connection failures
    #

    '''
    Test if the correct error message gets printed by the check method on connection failure.
    Connection failure does not need to be mocked. We really try to connect here.
    '''
    @unittest.skipIf(globalvar.QUICK_ONLY, '(this test is slow)')
    def test_run_check_connection_failed(self):

        print(self.slow_message)

        # Test variables:
        rabbit1 = dict(
            user = 'johndoe',
            password = 'abc123yx',
            url = 'this.is.my.only.host')
        testconnector = TESTHELPERS.get_connector(messaging_service_credentials=[rabbit1])

        # Run code to be tested and capture stout:
        with tests.utils.captureconsoleoutput.captured_output() as (out, err):
            returned_msg = esgfpid.check.check_pid_queue_availability(
                connector = testconnector,
                print_to_console = True,
                print_success_to_console = True
            )

        # Expected message:
        expected_message = tests.resources.error_messages.expected_message_connection_failed_only1
        expected_return = tests.resources.error_messages.expected_message_connection_failed_only2

        # Check result:
        output = out.getvalue().strip()
        self.assertEqual(expected_message, output,
            'Wrong stdout message.\n\nWe expected:\n\n%s\n\nWe got:\n\n%s\n' % (expected_message, output))
        self.assertEqual(expected_return, returned_msg,
            'Wrong return message.\n\nWe expected:\n\n%s\n\nWe got:\n\n%s\n' % (expected_return, returned_msg))

    @mock.patch('esgfpid.check.RabbitChecker._RabbitChecker__pika_blocking_connection')
    def test_run_check_connection_failed_unknown(self, connection_patch):
        '''
        Test if the correct error message gets printed by the check method on connection failure.
        Connection failure does not need to be mocked. We really try to connect here.
        '''

        # Define the replacement for the patched method
        connection_patch.return_value = None

        # Test variables:
        rabbit1 = dict(
            user = 'johndoe',
            password = 'abc123yx',
            url = 'this.is.my.only.host')
        testconnector = TESTHELPERS.get_connector(messaging_service_credentials=[rabbit1])

        # Run code to be tested and capture stout:
        with tests.utils.captureconsoleoutput.captured_output() as (out, err):
            returned_msg = esgfpid.check.check_pid_queue_availability(
                connector = testconnector,
                print_to_console = True,
                print_success_to_console = True
            )


        # Expected message:
        expected_message = tests.resources.error_messages.expected_message_connection_failed_unknown1
        expected_return = tests.resources.error_messages.expected_message_connection_failed_unknown2

        
        # Check result:
        output = out.getvalue().strip()
        self.assertEqual(expected_message, output,
            'Wrong stdout message.\n\nWe expected:\n\n%s\n\nWe got:\n\n%s\n' % (expected_message, output))
        self.assertEqual(expected_return, returned_msg,
            'Wrong return message.\n\nWe expected:\n\n%s\n\nWe got:\n\n%s\n' % (expected_return, returned_msg))



    '''Test if the correct error message gets printed by the check method on connection failure.'''
    @unittest.skipIf(globalvar.QUICK_ONLY, '(this test is slow)')
    def test_run_check_connection_failed_several_urls(self):

        print(self.slow_message)

        # Test variables:
        #messaging_service_urls = ['tomato.salad-with-spam.fr', 'mystery-tour.uk']
        #messaging_service_url_preferred = 'this.is.my.favourite.host'
        user = 'johndoe'
        pw = 'abc123yx'
        rabbit1 = dict(
            user = user,
            password = pw,
            url = 'this.is.my.favourite.host',
            priority=1)
        rabbit2 = dict(
            user = user,
            password = pw,
            priority=2,
            url = 'mystery-tour.uk')
        rabbit3 = dict(
            user = user,
            password = pw,
            priority=3,
            url = 'tomato.salad-with-spam.fr')
        testconnector = TESTHELPERS.get_connector(messaging_service_credentials=[rabbit1,rabbit2,rabbit3])

        # Run code to be tested and capture stout:
        with tests.utils.captureconsoleoutput.captured_output() as (out, err):
            returned_msg = esgfpid.check.check_pid_queue_availability(
                connector = testconnector,
                print_to_console = True,
                print_success_to_console = True
            )

        # Expected message:
        expected_message = tests.resources.error_messages.expected_message_connection_failed_several1
        expected_return = tests.resources.error_messages.expected_message_connection_failed_several2

        # Check result:
        output = out.getvalue().strip()
        self.assertEqual(expected_message, output,
            'Wrong return message.\n\nWe expected:\n\n%s\n\nWe got:\n\n%s\n' % (expected_message, output))
        self.assertEqual(expected_return, returned_msg,
            'Wrong stdout message.\n\nWe expected:\n\n%s\n\nWe got:\n\n%s\n' % (expected_return, returned_msg))

    @mock.patch('esgfpid.check.RabbitChecker._RabbitChecker__pika_blocking_connection')
    def test_run_check_first_connection_failed(self, connection_patch):
        '''
        Test if the correct error message gets printed by the check method on connection failure.
        Connection failure does not need to be mocked. We really try to connect here.
        '''

        # Test variables:
        #messaging_service_url_preferred = 'this.is.my.favourite.host'
        #messaging_service_urls =  ['tomato.salad-with-spam.fr', 'mystery-tour.uk']
        user = 'johndoe'
        pw = 'abc123yx'
        rabbit1 = dict(
            user = user,
            password = pw,
            priority=1,
            url = 'this.is.my.favourite.host')
        rabbit2 = dict(
            user = user,
            password = pw,
            priority=3,
            url = 'tomato.salad-with-spam.fr')
        rabbit3 = dict(
            user = user,
            password = pw,
            priority=2,
            url = 'mystery-tour.uk')
        testconnector = TESTHELPERS.get_connector(messaging_service_credentials=[rabbit1,rabbit2,rabbit3])

        # Define the replacement for the patched method:
        def different_mock_response_depending_on_host(params):
            if params.host == 'tomato.salad-with-spam.fr':
                return tests.resources.pikamock.MockPikaBlockingConnection(mock.MagicMock())
            else:
                raise pika.exceptions.ConnectionClosed(reply_code=0, reply_text='Connection Closed')      
        connection_patch.side_effect = different_mock_response_depending_on_host

        # Run code to be tested and capture stout:
        with tests.utils.captureconsoleoutput.captured_output() as (out, err):
            returned_msg = esgfpid.check.check_pid_queue_availability(
                connector = testconnector,
                print_to_console = True,
                print_success_to_console = True
            )

        # Expected message:
        expected_message = tests.resources.error_messages.expected_message_first_connection_failed_then_ok1
        expected_return = tests.resources.error_messages.expected_message_first_connection_failed_then_ok2

        # Check result:
        output = out.getvalue().strip()
        self.assertEqual(expected_message, output,
            'Wrong stdout message.\n\nWe expected:\n\n%s\n\nWe got:\n\n%s\n' % (expected_message, output))
        self.assertEqual(expected_return, returned_msg,
            'Wrong return message.\n\nWe expected:\n\n%s\n\nWe got:\n\n%s\n' % (expected_return, returned_msg))


    #
    # Authentication failure
    #

    @mock.patch('esgfpid.check.RabbitChecker._RabbitChecker__pika_blocking_connection')
    def test_run_check_auth_error(self, connection_patch):

        # Define the replacement for the patched method:
        mock_response = tests.resources.pikamock.MockPikaBlockingConnection(mock.MagicMock())
        connection_patch.return_value = mock_response
        connection_patch.side_effect = pikaexceptions.ProbableAuthenticationError

        # Test variables:
        rabbit1 = dict(
            user = 'johndoe',
            password = 'abc123yx',
            url = 'this.is.my.only.host')
        testconnector = TESTHELPERS.get_connector(messaging_service_credentials=[rabbit1])

        # Run code to be tested and capture stout:
        with tests.utils.captureconsoleoutput.captured_output() as (out, err):
            returned_msg =  esgfpid.check.check_pid_queue_availability(
                connector = testconnector,
                print_to_console = True,
                print_success_to_console = True
            )

        # Expected message:
        expected_message = tests.resources.error_messages.expected_message_authentication_failed_only1
        expected_return = tests.resources.error_messages.expected_message_authentication_failed_only2

        # Check result:
        output = out.getvalue().strip()
        self.assertEqual(expected_message, output,
            'Wrong stdout message.\n\nWe expected:\n\n%s\n\nWe got:\n\n%s\n' % (expected_message, output))
        self.assertEqual(expected_return, returned_msg,
            'Wrong return message.\n\nWe expected:\n\n%s\n\nWe got:\n\n%s\n' % (expected_return, returned_msg))


    @mock.patch('esgfpid.check.RabbitChecker._RabbitChecker__pika_blocking_connection')
    def test_run_check_auth_error_three_urls(self, connection_patch):

        # Define the replacement for the patched method:
        mock_response = tests.resources.pikamock.MockPikaBlockingConnection(mock.MagicMock())
        connection_patch.return_value = mock_response
        connection_patch.side_effect = pikaexceptions.ProbableAuthenticationError

        # Test variables:
        #messaging_service_urls =  ['tomato.salad-with-spam.fr', 'mystery-tour.uk']
        #messaging_service_url_preferred = 'this.is.my.favourite.host'
        user = 'johndoe'
        pw = 'abc123yx'
        rabbit1 = dict(
            user = user,
            password = pw,
            url = 'this.is.my.favourite.host',
            priority=1)
        rabbit2 = dict(
            user = user,
            password = pw,
            priority=3,
            url = 'tomato.salad-with-spam.fr')
        rabbit3 = dict(
            user = user,
            password = pw,
            priority=2,
            url = 'mystery-tour.uk')
        testconnector = TESTHELPERS.get_connector(messaging_service_credentials=[rabbit1,rabbit2,rabbit3])

        # Run code to be tested and capture stout:
        with tests.utils.captureconsoleoutput.captured_output() as (out, err):
            returned_msg = esgfpid.check.check_pid_queue_availability(
                connector = testconnector,
                print_to_console = True,
                print_success_to_console = True
            )

        # Expected message:
        expected_message = tests.resources.error_messages.expected_message_authentication_failed_several1
        expected_return = tests.resources.error_messages.expected_message_authentication_failed_several2

        # Check result:
        output = out.getvalue().strip()
        self.assertEqual(expected_message, output,
            'Wrong stdout message.\n\nWe expected:\n\n%s\n\nWe got:\n\n%s\n' % (expected_message, output))
        self.assertEqual(expected_return, returned_msg,
            'Wrong return message.\n\nWe expected:\n\n%s\n\nWe got:\n\n%s\n' % (expected_return, returned_msg))

    #
    # Channel error
    #

    @mock.patch('esgfpid.check.RabbitChecker._RabbitChecker__pika_blocking_connection')
    def test_run_check_channel_error(self, connection_patch):

        # Define the replacement for the patched method:
        mock_response = tests.resources.pikamock.MockPikaBlockingConnection(mock.MagicMock())
        mock_response.raise_channel_closed = True
        connection_patch.return_value = mock_response

        # Test variables:
        rabbit1 = dict(
            user = 'johndoe',
            password = 'abc123yx',
            url = 'this.is.my.only.host')
        testconnector = TESTHELPERS.get_connector(messaging_service_credentials=[rabbit1])

        # Run code to be tested and capture stout:
        with tests.utils.captureconsoleoutput.captured_output() as (out, err):
            returned_msg = esgfpid.check.check_pid_queue_availability(
                connector = testconnector,
                print_to_console = True,
                print_success_to_console = True
            )

        # Expected message:
        expected_message = tests.resources.error_messages.expected_message_channel_closed_only1
        expected_return = tests.resources.error_messages.expected_message_channel_closed_only2

        # Check result:
        output = out.getvalue().strip()
        self.assertEqual(expected_message, output,
            'Wrong stdout message.\n\nWe expected:\n\n%s\n\nWe got:\n\n%s\n' % (expected_message, output))
        self.assertEqual(expected_return, returned_msg,
            'Wrong return message.\n\nWe expected:\n\n%s\n\nWe got:\n\n%s\n' % (expected_return, returned_msg))

    @mock.patch('esgfpid.check.RabbitChecker._RabbitChecker__pika_blocking_connection')
    def test_run_check_channel_error_first(self, connection_patch):
        '''
        Test if the correct error message gets printed by the check method on connection failure.
        Connection failure does not need to be mocked. We really try to connect here.
        '''

        # Test variables:
        #messaging_service_url_preferred = 'this.is.my.favourite.host'
        #messaging_service_urls =  ['tomato.salad-with-spam.fr', 'mystery-tour.uk']
        user = 'johndoe'
        pw = 'abc123yx'
        rabbit1 = dict(
            user = user,
            password = pw,
            priority=1,
            url = 'this.is.my.favourite.host')
        rabbit2 = dict(
            user = user,
            password = pw,
            priority=3,
            url = 'tomato.salad-with-spam.fr')
        rabbit3 = dict(
            user = user,
            password = pw,
            priority=2,
            url = 'mystery-tour.uk')
        testconnector = TESTHELPERS.get_connector(messaging_service_credentials=[rabbit1,rabbit2,rabbit3])

        # Define the replacement for the patched method:
        def different_mock_response_depending_on_host(params):
            conn = tests.resources.pikamock.MockPikaBlockingConnection(mock.MagicMock())
            if params.host == 'tomato.salad-with-spam.fr':
                conn.raise_channel_closed = False
            else:
                conn.raise_channel_closed = True
            return conn
        connection_patch.side_effect = different_mock_response_depending_on_host

        # Run code to be tested and capture stout:
        with tests.utils.captureconsoleoutput.captured_output() as (out, err):
            returned_msg = esgfpid.check.check_pid_queue_availability(
                connector = testconnector,
                print_to_console = True,
                print_success_to_console = True
            )

        # Expected message:
        expected_message = tests.resources.error_messages.expected_message_first_channel_closed_then_ok1
        expected_return = tests.resources.error_messages.expected_message_first_channel_closed_then_ok2

        # Check result:
        output = out.getvalue().strip()
        self.assertEqual(expected_message, output,
            'Wrong stdout message.\n\nWe expected:\n\n%s\n\nWe got:\n\n%s\n' % (expected_message, output))
        self.assertEqual(expected_return, returned_msg,
            'Wrong return message.\n\nWe expected:\n\n%s\n\nWe got:\n\n%s\n' % (expected_return, returned_msg))





import unittest
import mock
import logging
import json
import requests
import esgfpid.rabbit.asynchronous
from esgfpid.rabbit.asynchronous.exceptions import OperationNotAllowed
import tests.mocks.responsemock
import tests.mocks.pikamock
import mock
import time
import os

# Logging:
LOGGER = logging.getLogger(__name__)
LOGGER.addHandler(logging.NullHandler())

import globalvar
if globalvar.QUICK_ONLY:
    print('Skipping slow tests in module "%s".' % __name__)

class RabbitAsynModuleTestCase(unittest.TestCase):

    slow_message = '\nRunning a slow test (avoid by using -ls flag).'

    def setUp(self):
        LOGGER.info('######## Next test (%s) ##########', __name__)
        self.testrabbit = None
        self.thread = None

    def tearDown(self):
        LOGGER.info('#############################')

    def make_nodemanager(self):
        # Make a node manager with three nodes:
        node_manager = esgfpid.rabbit.nodemanager.NodeManager()
        node_manager.add_trusted_node(
            username='userfoo',
            password='passwordfoo',
            host='first',
            exchange_name='foo_exchange'
        )
        node_manager.add_open_node(
            username='userfoo',
            password='passwordfoo',
            host='foo',
            exchange_name='foo_exchange'
        )
        node_manager.add_open_node(
            username='userfoo',
            password='passwordfoo',
            host='bar',
            exchange_name='foo_exchange'
        )
        return node_manager

    def make_testrabbit(self):

        node_manager = self.make_nodemanager()
        testrabbit = esgfpid.rabbit.asynchronous.AsynchronousRabbitConnector(node_manager)
        conn = tests.mocks.pikamock.MockPikaSelectConnection()

        self.rabbit = testrabbit
        self.thread = self.rabbit._AsynchronousRabbitConnector__thread
        self.builder = self.thread._RabbitThread__builder
        self.connection = conn
        callback = None
        self.channel = conn.channel(callback)


    # Actual tests:


    '''
    Test the constructor.
    '''
    def test_init_ok(self):

        # Test variables:
        node_manager = esgfpid.rabbit.nodemanager.NodeManager()

        # Run code to be tested:
        testrabbit = esgfpid.rabbit.asynchronous.AsynchronousRabbitConnector(node_manager)

        # Check result:
        self.assertIsInstance(testrabbit, esgfpid.rabbit.asynchronous.AsynchronousRabbitConnector, 'Constructor fail.')
        self.assertTrue(testrabbit._AsynchronousRabbitConnector__not_started_yet)


    #
    # Sending messages
    #

    '''
    Test if we can send messages before starting the thread.
    We shouldn't!
    '''
    def test_send_message_not_started_yet(self):
        self.make_testrabbit()

        self.assertTrue(self.rabbit._AsynchronousRabbitConnector__not_started_yet)

        with self.assertRaises(OperationNotAllowed):
            self.rabbit.send_message_to_queue('message-foo')

        with self.assertRaises(OperationNotAllowed):
                self.rabbit.send_many_messages_to_queue(['a','b','c'])

    @unittest.skipIf(globalvar.QUICK_ONLY, '(this test is slow)')
    @mock.patch('esgfpid.rabbit.asynchronous.thread_builder.ConnectionBuilder._ConnectionBuilder__please_open_connection')
    def test_send_message_ok_no_confirm(self, open_connection_patch):

        print('\nThis test may take some seconds, as it waits for confirms that can not come...')
        print(self.slow_message)

        # Patch the method that tries to open a conneciton to RabbitMQ:
        open_connection_patch = mock.MagicMock()

        # Make a test object:
        self.make_testrabbit()

        # Start the thread. This would normally trigger
        # the connection, but we patched that part, and
        # have to  it manually:
        self.rabbit.start_rabbit_thread()
        # This would usually be done by Rabbit via callback,
        # when start_rabbit_thread is called:
        self.thread._connection = self.connection
        self.builder.on_connection_open(None)
        self.builder.on_channel_open(self.channel)

        # Now run the code to be tested and close the rabbit:
        self.rabbit.send_message_to_queue({"bla":"bla"})
        self.rabbit.send_many_messages_to_queue([{"bla":"bla"},{"bla":"bla"},{"bla":"bla"}])
        self.rabbit.finish_rabbit_thread()

        # Check result: Are 4 messages published?
        self.assertEqual(self.channel.publish_counter, 4)

    @mock.patch('esgfpid.rabbit.asynchronous.thread_builder.ConnectionBuilder._ConnectionBuilder__please_open_connection')
    def test_finish_ok(self, open_connection_patch):

        # Patch the method that tries to open a conneciton to RabbitMQ:
        open_connection_patch = mock.MagicMock()

        # Make a test object:
        self.make_testrabbit()

        # Start the thread. This would normally trigger
        # the connection, but we patched that part, and
        # have to  it manually:
        self.rabbit.start_rabbit_thread()
        # This would usually be done by Rabbit via callback,
        # when start_rabbit_thread is called:
        self.thread._connection = self.connection
        self.builder.on_connection_open(None)
        self.builder.on_channel_open(self.channel)

        # Now run the code to be tested and close the rabbit:
        self.rabbit.finish_rabbit_thread()

        # Check result
        self.assertTrue(self.builder.statemachine.is_PERMANENTLY_UNAVAILABLE())
        self.assertTrue(self.builder.statemachine.get_detail_closed_by_publisher())
        self.assertTrue(self.thread._connection.is_closed)
        self.assertFalse(self.thread._connection.is_open)

    @mock.patch('esgfpid.rabbit.asynchronous.thread_builder.ConnectionBuilder._ConnectionBuilder__please_open_connection')
    def test_force_finish_ok(self, open_connection_patch):

        # Patch the method that tries to open a conneciton to RabbitMQ:
        open_connection_patch = mock.MagicMock()

        # Make a test object:
        self.make_testrabbit()

        # Start the thread. This would normally trigger
        # the connection, but we patched that part, and
        # have to  it manually:
        self.rabbit.start_rabbit_thread()
        # This would usually be done by Rabbit via callback,
        # when start_rabbit_thread is called:
        self.thread._connection = self.connection
        self.builder.on_connection_open(None)
        self.builder.on_channel_open(self.channel)

        # Now run the code to be tested and close the rabbit:
        self.rabbit.force_finish_rabbit_thread()

        # Check result
        self.assertTrue(self.builder.statemachine.is_PERMANENTLY_UNAVAILABLE())
        self.assertTrue(self.builder.statemachine.get_detail_closed_by_publisher())
        self.assertTrue(self.thread._connection.is_closed)
        self.assertFalse(self.thread._connection.is_open)



import unittest
import mock
import logging
import json
import sys
import datetime
import uuid
import os
import esgfpid

# Logging
LOGGER = logging.getLogger(__name__)
LOGGER.addHandler(logging.NullHandler())

# Test resources:
from resources.TESTVALUES import *


'''
Unit tests for esgfpid.utils.

This module does not need any other module to
function, so we don't need to use or to mock any
other objects. 
'''
class UtilsTestCase(unittest.TestCase):

    def setUp(self):
        LOGGER.info('######## Next test (%s) ##########', __name__)

    def tearDown(self):
        LOGGER.info('#############################')

    #
    # handleutils
    #

    def test_make_handle_from_drsid_and_versionnumber_ok(self):

        received_handle = esgfpid.utils.make_handle_from_drsid_and_versionnumber(
            drs_id=DRS_ID,
            version_number=DS_VERSION,
            prefix=PREFIX_NO_HDL
        )

        expected_handle = DATASETHANDLE_HDL
        self.assertEqual(received_handle, expected_handle)

    #
    # argsutils
    #


    def test_check_presence_of_mandatory_args_ok(self):

        # Test variables
        args = {}
        args['mand1']='val1'
        args['mand2']='val2'
        args['mand3']='val3'
        args['opt1']='val1'
        args['opt2']='val2'
        args['opt3']='val3'
        mandatory_args = ['mand1','mand2','mand3']
  
        # Run code to be tested:

        all_mandatory_args_ok = esgfpid.utils.check_presence_of_mandatory_args(args, mandatory_args)

        # Check result:
        self.assertTrue(all_mandatory_args_ok, 'Checking mandatory args did not work.')

    def test_check_presence_of_mandatory_args_one_missing(self):

        # Test variables
        args = {}
        args['mand1']='val1'
        args['mand2']='val2'
        args['opt1']='val1'
        args['opt2']='val2'
        args['opt3']='val3'
        mandatory_args = ['mand1','mand2','mand3']
  
        # Run code to be tested and check exception:
        with self.assertRaises(esgfpid.exceptions.ArgumentError)as raised:
            all_mandatory_args_ok = esgfpid.utils.check_presence_of_mandatory_args(args, mandatory_args)
        expected_message = 'The arguments that were passed are not ok: Missing mandatory arguments: mand3'
        self.assertIn(expected_message, repr(raised.exception),
            'Unexpected error message:\n%s\nExpected:\n%s' % (repr(raised.exception), expected_message))

    def test_check_presence_of_mandatory_args_several_missing(self):

        # Test variables
        args = {}
        args['mand1']='val1'
        args['opt1']='val1'
        mandatory_args = ['mand1','mand2','mand3']
  
        # Run code to be tested and check exception:
        with self.assertRaises(esgfpid.exceptions.ArgumentError)as raised:
            all_mandatory_args_ok = esgfpid.utils.check_presence_of_mandatory_args(args, mandatory_args)
        expected_message = 'The arguments that were passed are not ok: Missing mandatory arguments: mand2, mand3'
        self.assertIn(expected_message, repr(raised.exception),
            'Unexpected error message:\n%s\nExpected:\n%s' % (repr(raised.exception), expected_message))

    def test_find_additional_args_found_ok(self):

        # Test variables:
        mandatory_args = ['mand1','mand2','mand3']
        optional_args = ['opt1','opt2','opt3']
        all_args = dict(
            mand1=100,
            opt2=100,
            another=10000,
            mand3=100,
            opt1=100,
            mand2=100,
            opt3=100
        )

        # Run code to be tested:
        received = esgfpid.utils.find_additional_args(all_args, (mandatory_args+optional_args))

        # Check result:
        expected = {'another':10000}
        self.assertEqual(received, expected)

    def test_find_additional_args_none_found(self):

        # Test variables:
        mandatory_args = ['mand1','mand2','mand3']
        optional_args = ['opt1','opt2','opt3']
        all_args = dict(
            mand1=100,
            opt2=100,
            mand3=100,
            opt1=100,
            mand2=100,
            opt3=100
        )

        # Run code to be tested:
        received = esgfpid.utils.find_additional_args(all_args, (mandatory_args+optional_args))

        # Check result:
        expected = {}
        self.assertEqual(received, expected)

    def test_check_noneness_of_mandatory_args_none(self):
        mandatory_args = ['mand1','mand2']

        all_args = dict(
            mand1=100,
            opt2=100,
            another=10000,
            mand2=None,
        )

        with self.assertRaises(esgfpid.exceptions.ArgumentError) as raised:
            esgfpid.utils.check_noneness_of_mandatory_args(all_args, mandatory_args)
        self.assertIn('These arguments are None: mand2', repr(raised.exception))

    def test_check_noneness_of_mandatory_args_ok(self):
        mandatory_args = ['mand1','mand2']

        all_args = dict(
            mand1=100,
            opt2=100,
            another=10000,
            mand2=100,
        )

        success = esgfpid.utils.check_noneness_of_mandatory_args(all_args, mandatory_args)
        self.assertTrue(success)

    def test_add_missing_optional_args_with_value_none_ok(self): 
        mandatory_args = ['mand1','mand2']
        optional_args = ['opt1','opt2']

        all_args = dict(
            mand1=100,
            opt2=100,
            another=10000,
            mand2=100,
        )
        esgfpid.utils.add_missing_optional_args_with_value_none(all_args, optional_args)
        self.assertIn('opt1', all_args)
        self.assertIn('opt2', all_args)
        self.assertTrue(all_args['opt1'] is None)
        self.assertTrue(all_args['opt2'] is not None)


    def test_add_missing_optional_args_with_value_none_not_needed(self): 
        mandatory_args = ['mand1','mand2']
        optional_args = ['opt1','opt2']

        all_args = dict(
            mand1=100,
            opt1=100,
            opt2=100,
            another=10000,
            mand2=100,
        )
        esgfpid.utils.add_missing_optional_args_with_value_none(all_args, optional_args)
        self.assertIn('opt1', all_args)
        self.assertIn('opt2', all_args)
        self.assertTrue(all_args['opt1'] is not None)
        self.assertTrue(all_args['opt2'] is not None)

    #
    # timeutils
    #

    def test_get_now_utc(self):
        received = esgfpid.utils.get_now_utc()
        self.assertIsInstance(received, datetime.datetime, 'Unexpected time type: %s' % type(received))

    def test_get_now_utc_as_string(self):
        received = esgfpid.utils.get_now_utc_as_formatted_string() #2016-07-07T15:25:18.258224+00:00
        self.assertIn('%s-' % datetime.datetime.now().year, received, 'Unexpected time: %s' % received)
        self.assertIn('T', received, 'Unexpected time: %s' % received)
        self.assertIn('+00:00', received, 'Unexpected time: %s' % received)

    #
    # miscutils - get_boolean
    #

    def test_get_boolean_true(self):
        self.assertTrue(esgfpid.utils.get_boolean(True))
        self.assertTrue(esgfpid.utils.get_boolean('True'))
        self.assertTrue(esgfpid.utils.get_boolean('true'))

    def test_get_boolean_false(self):
        self.assertFalse(esgfpid.utils.get_boolean(False))
        self.assertFalse(esgfpid.utils.get_boolean('False'))
        self.assertFalse(esgfpid.utils.get_boolean('false'))

    def test_get_boolean_string_wrong(self):
        with self.assertRaises(ValueError):
            self.assertFalse(esgfpid.utils.get_boolean('maybe'))

    def test_get_boolean_number_wrong(self):
        with self.assertRaises(ValueError):
            self.assertFalse(esgfpid.utils.get_boolean(123))

    #
    # miscutils - directory helper
    #

    def test_ensure_dir_exists_alreadythere(self):
        ''' The method should not do anything if the directory exists already.'''

        # Make a test dir:
        workingdir = os.getcwd()
        mydir = workingdir+'/'+str(uuid.uuid4())
        os.mkdir(mydir)
        if not os.path.isdir(mydir):
            raise ValueError('We just created this dir, it should exist: %s' % mydir)

        # Run method:
        esgfpid.utils.ensure_directory_exists(mydir)

        # Check result:
        self.assertTrue(os.path.isdir(mydir), 'Directory does not exist. This is strange.')

        # Delete again:
        os.rmdir(mydir)
        if os.path.isdir(mydir):
          raise ValueError('Could not remove test dir after testing: %s' % mydir)

    def test_ensure_dir_exists_notthereyet(self):
        ''' The method should create the directory, as it does not exist.'''

        # Make a name of test dir:
        workingdir = os.getcwd()
        mydir = workingdir+'/'+str(uuid.uuid4())
        if os.path.isdir(mydir):
            raise ValueError('This directory should not exist: %s' % mydir)

        # Run method:
        esgfpid.utils.ensure_directory_exists(mydir)

        # Check result:
        self.assertTrue(os.path.isdir(mydir), 'Directory does not exist. This is strange.')

        # Delete again:
        os.rmdir(mydir)
        if os.path.isdir(mydir):
          raise ValueError('Could not remove test dir after testing: %s' % mydir)




import logging
import esgfpid.rabbit
import esgfpid.solr
import esgfpid.utils

LOGGER = logging.getLogger(__name__)
LOGGER.addHandler(logging.NullHandler())

'''
This class is responsible for communication with the
outside world, i.e. with solr and/or RabbitMQ.

Solr is needed for checking consistency, but these
checks are usually disabled.

The communication with RabbitMQ is crucial for the
library.

Everything that the library needs to know from solr
or RabbitMQ is done via this coupling module, which
basically just forwards all request to the responsible
module.
'''
class Coupler(object):

    '''
    Create a Coupler instance.

    :param messaging_service_credentials: Mandatory.
    :param messaging_service_exchange_name: Mandatory.
    :param message_service_synchronous: Mandatory. Boolean.
    :param test_publication: Mandatory. Boolean.

    :param solr_switched_off: Mandatory. Boolean.
    :param solr_url: Mandatory. May be None if switched off.
    :param handle_prefix: Mandatory. May be None if switched off.
    :param solr_https_verify: Mandatory. May be None if switched off.
    :param disable_insecure_request_warning: Mandatory. May be None if switched off.

    '''
    def __init__(self, **args):
        self.args = args # only for unit testing
        self.__create_message_sender(args)
        self.__create_solr_sender(args)     

    def __create_message_sender(self, args):
        self.__complete_credentials_for_open_nodes(args)
        self.__rabbit_message_sender = esgfpid.rabbit.RabbitMessageSender(
            exchange_name=args['messaging_service_exchange_name'],
            credentials=args['messaging_service_credentials'],
            test_publication=args['test_publication'],
            is_synchronous_mode=args['message_service_synchronous']
        )

    def __complete_credentials_for_open_nodes(self, args):
        credentials = args['messaging_service_credentials']
        for cred in credentials:
            if 'password' not in cred:
                cred['password'] = 'jzlnL78ZpExV#_QHz'

    def __create_solr_sender(self, args):
        self.__solr_sender = esgfpid.solr.SolrInteractor(
            solr_url=args['solr_url'],
            prefix=args['handle_prefix'],
            https_verify=args['solr_https_verify'],
            switched_off=args['solr_switched_off'],
            disable_insecure_request_warning = ['disable_insecure_request_warning']
        )

    ### Communications with rabbit

    '''
    Please see documentation of rabbit module (:func:`~rabbit.RabbitMessageSender.send_message_to_queue`).
    '''
    def send_message_to_queue(self, message):
        self.__rabbit_message_sender.send_message_to_queue(message)

    ### For synchronous

    '''
    Please see documentation of rabbit module (:func:`~rabbit.RabbitMessageSender.open_rabbit_connection`).
    '''
    def start_rabbit_business(self):
        self.__rabbit_message_sender.open_rabbit_connection()

    '''
    Please see documentation of rabbit module (:func:`~rabbit.RabbitMessageSender.close_rabbit_connection`).
    '''
    def done_with_rabbit_business(self):
        self.__rabbit_message_sender.close_rabbit_connection()

    ### For asynchronous

    '''
    Please see documentation of rabbit module (:func:`~rabbit.RabbitMessageSender.start`).
    '''
    def start_rabbit_connection(self):
        self.__rabbit_message_sender.start()

    '''
    Please see documentation of rabbit module (:func:`~rabbit.RabbitMessageSender.finish`).
    '''
    def finish_rabbit_connection(self):
        self.__rabbit_message_sender.finish()

    '''
    Please see documentation of rabbit module (:func:`~rabbit.RabbitMessageSender.force_finish`).
    '''
    def force_finish_rabbit_connection(self):
        self.__rabbit_message_sender.force_finish()

    ### Communications with solr

    '''
    Please see documentation of solr module (:func:`~solr.SolrInteractor.retrieve_datasethandles_or_versionnumbers_of_allversions`).
    '''
    def retrieve_datasethandles_or_versionnumbers_of_allversions(self, **args):
        mandatory_args = ['drs_id']
        esgfpid.utils.check_presence_of_mandatory_args(args, mandatory_args)
        esgfpid.utils.check_noneness_of_mandatory_args(args, mandatory_args)

        result_dict = self.__solr_sender.retrieve_datasethandles_or_versionnumbers_of_allversions(
            drs_id=args['drs_id'],
        )
        return result_dict

    '''
    Please see documentation of solr module (:func:`~solr.SolrInteractor.retrieve_file_handles_of_same_dataset`).
    '''
    def retrieve_file_handles_of_same_dataset(self, **args):
        mandatory_args = ['drs_id', 'data_node', 'version_number']
        esgfpid.utils.check_presence_of_mandatory_args(args, mandatory_args)
        esgfpid.utils.check_noneness_of_mandatory_args(args, mandatory_args)

        result_dict = self.__solr_sender.retrieve_file_handles_of_same_dataset(
            drs_id=args['drs_id'],
            data_node=args['data_node'],
            version_number=args['data_node']
        )
        return result_dict

    '''
    Please see documentation of solr module (:func:`~solr.SolrInteractor.is_switched_off`).
    '''
    def is_solr_switched_off(self):
        return self.__solr_sender.is_switched_off()


import logging
import pika
import time
import copy
import datetime
from esgfpid.utils import get_now_utc_as_formatted_string as get_now_utc_as_formatted_string
import esgfpid.defaults as defaults
from esgfpid.utils import loginfo, logdebug, logtrace, logerror, logwarn, log_every_x_times
from ..exceptions import PIDServerException

LOGGER = logging.getLogger(__name__)
LOGGER.addHandler(logging.NullHandler())

'''

If the module fails connecting to a RabbitMQ node (on_connection_error),
or if the connection if interrupted (on_connection_closed),
it immediately tries connecting to the next RabbitMQ node.

If all hosts have been tried, the module starts over again, but waits some
seconds before that.

There is a maximum number of times that this is tried before
giving up Permanently.

'''
class ConnectionBuilder(object):
    
    def __init__(self, thread, statemachine, confirmer, returnhandler, shutter, nodemanager):
        self.thread = thread
        self.statemachine = statemachine

        '''
        We need to pass the "confirmer.on_delivery_confirmation()" callback to
        RabbitMQ's channel.'''
        self.confirmer = confirmer
        
        '''
        We need to pass the "returnhandler.on_message_not_accepted()"" callback
        to RabbitMQ's channel as "on_return_callback" '''
        self.returnhandler = returnhandler

        '''
        We need this to be able to trigger all the closing mechanisms 
        in case the module should close down as soon it was opened, i.e.
        if the close-command was issued while the connection was still
        building up.
        '''
        self.shutter = shutter

        '''
        The node manager keeps all the info about the RabbitMQ nodes,
        e.g. URLs, usernames, passwords.
        '''
        self.__node_manager = nodemanager

        '''
        To count how many times we have tried to reconnect the set of
        RabbitMQ hosts.
        '''
        self.__reconnect_counter = 0
        self.__backup_reconnect_counter = 0

        '''
        To see how many times we should try reconnecting to the set 
        of RabbitMQ hosts. Note that if there is 3 hosts, and we try 2
        times, this means 6 connection tries in total.
        '''
        self.__max_reconnection_tries = defaults.RABBIT_RECONNECTION_MAX_TRIES

        '''
        How many seconds to wait before reconnecting after having tried
        all hosts. (There is no waiting time trying to connect to a different
        host after one fails).
        '''
        self.__wait_seconds_before_reconnect = defaults.RABBIT_RECONNECTION_SECONDS

        '''
        To see how much time it takes to connect. Once a connection is
        established or failed, we print the time delta to logs.
        '''
        self.__start_connect_time = None

        '''
        Name of the fallback exchange to try if the normal exchange
        is not found.
        '''
        self.__fallback_exchange_name = defaults.RABBIT_FALLBACK_EXCHANGE_NAME

        '''
        Collect the connection errors for the hosts for telling the user.
        '''
        self.__connection_errors = {}

    ####################
    ### Start ioloop ###
    ####################

    '''
    Entry point. Called once to trigger the whole
    (re) connection process. Called from run method of the rabbit thread.
    '''
    def first_connection(self):
        logdebug(LOGGER, 'Trigger connection to rabbit...')
        self.__trigger_connection_to_rabbit_etc()
        logdebug(LOGGER, 'Trigger connection to rabbit... done.')
        logdebug(LOGGER, 'Start waiting for events...')
        self.__start_waiting_for_events()
        logtrace(LOGGER, 'Had started waiting for events, but stopped.')
    
    def __start_waiting_for_events(self):
        '''
        This waits until the whole chain of callback methods triggered by
        "trigger_connection_to_rabbit_etc()" has finished, and then starts 
        waiting for publications.
        This is done by starting the ioloop.

        Note: In the pika usage example, these things are both called inside the run()
        method, so I wonder if this check-and-wait here is necessary. Maybe not.
        But the usage example does not implement a Thread, so it probably blocks during
        the opening of the connection. Here, as it is a different thread, the run()
        might get called before the __init__ has finished? I'd rather stay on the
        safe side, as my experience of threading in Python is limited.
        '''

        # Start ioloop if connection object ready:
        if self.thread._connection is not None:
            try:
                logdebug(LOGGER, 'Starting ioloop...')
                logtrace(LOGGER, 'ioloop is owned by connection %s...', self.thread._connection)

                # Tell the main thread that we're now open for events.
                # As soon as the thread._connection object is not None anymore, it
                # can receive events.
                self.thread.tell_publisher_to_stop_waiting_for_thread_to_accept_events() 
                self.thread.continue_gently_closing_if_applicable()
                self.thread._connection.ioloop.start()

            except PIDServerException as e:
                raise e

            # It seems that some connection problems do not cause
            # RabbitMQ to call any callback (on_connection_closed
            # or on_connection_error) - it just silently swallows the
            # problem.
            # So we need to manually trigger reconnection to the next
            # host here, which we do by manually calling the callback.
            # We start the ioloop, so it can handle the reconnection events,
            # or also receive events from the publisher in the meantime.


            except Exception as e:
                # This catches any error during connection startup and during the entire
                # time the ioloop runs, blocks and waits for events.

                time_passed = datetime.datetime.now() - self.__start_connect_time
                time_passed_seconds = time_passed.total_seconds()

                # Some pika errors:
                if isinstance(e, pika.exceptions.ProbableAuthenticationError):
                    errorname = self.__make_error_name(e, 'e.g. wrong user or password')

                elif isinstance(e, pika.exceptions.ProbableAccessDeniedError):
                    errorname = self.__make_error_name(e, 'e.g. wrong virtual host name')
                
                elif isinstance(e, pika.exceptions.IncompatibleProtocolError):
                    errorname = self.__make_error_name(e, 'e.g. trying TLS/SSL on wrong port')

                # Other errors:
                else:
                    errorname = self.__make_error_name(e)
                    logdebug(LOGGER, 'Unexpected error during event listener\'s lifetime (after %s seconds): %s', time_passed_seconds, errorname)

                # Now trigger reconnection:
                self.statemachine.set_to_waiting_to_be_available()
                self.on_connection_error(self.thread._connection, errorname)
                self.thread._connection.ioloop.start()
        
        else:
            # I'm quite sure that this cannot happen, as the connection object
            # is created in "trigger_connection_...()" and thus exists, no matter
            # if the actual connection to RabbitMQ succeeded (yet) or not.
            logdebug(LOGGER, 'This cannot happen: Connection object is not ready.')
            logerror(LOGGER, 'Cannot happen. Cannot properly start the thread. Connection object is not ready.')

    def __make_error_name(self, ex, custom_text=None):
        errorname = ex.__class__.__name__
        if not repr(ex) == '':
            errorname += ': '+repr(ex)
        if custom_text is not None:
            errorname += ' ('+custom_text+')'
        return errorname

    ########################################
    ### Chain of callback functions that ###
    ### connect to rabbit                ###
    ########################################

    def __trigger_connection_to_rabbit_etc(self):
        self.statemachine.set_to_waiting_to_be_available()
        self.__please_open_connection()

    ''' Asynchronous, waits for answer from RabbitMQ.'''
    def __please_open_connection(self):
        params = self.__node_manager.get_connection_parameters()
        self.__start_connect_time = datetime.datetime.now()
        logdebug(LOGGER, 'Connecting to RabbitMQ at %s... (%s)',
            params.host, get_now_utc_as_formatted_string())
        loginfo(LOGGER, 'Opening connection to RabbitMQ...')
        self.thread._connection = pika.SelectConnection(
            parameters=params,
            on_open_callback=self.on_connection_open,
            on_open_error_callback=self.on_connection_error,
            on_close_callback=self.on_connection_closed
            # Removed parameter, see https://github.com/pika/pika/issues/961
        )

    ''' Callback, called by RabbitMQ.'''
    def on_connection_open(self, unused_connection):
        logdebug(LOGGER, 'Opening connection... done.')
        loginfo(LOGGER, 'Connection to RabbitMQ at %s opened... (%s)',
            self.__node_manager.get_connection_parameters().host,
            get_now_utc_as_formatted_string())

        # Tell the main thread we're open for events now:
        # When the connection is open, the thread is ready to accept events.
        # Note: It was already ready when the connection object was created,
        # not just now that it's actually open. There was already a call to
        # "...stop_waiting..." in start_waiting_for_events(), which quite
        # certainly was carried out before this callback. So this call to
        # "...stop_waiting..." is likelily redundant!
        self.thread.tell_publisher_to_stop_waiting_for_thread_to_accept_events()
        self.__please_open_rabbit_channel()

    ''' Asynchronous, waits for answer from RabbitMQ.'''
    def __please_open_rabbit_channel(self):
        logdebug(LOGGER, 'Opening channel...')
        self.thread._connection.channel(on_open_callback=self.on_channel_open)

    ''' Callback, called by RabbitMQ. '''
    def on_channel_open(self, channel):
        time_passed = datetime.datetime.now() - self.__start_connect_time
        logdebug(LOGGER, 'Opening channel... done. Took %s seconds.' % time_passed.total_seconds())
        logtrace(LOGGER, 'Channel has number: %s.', channel.channel_number)
        self.thread._channel = channel
        self.__reset_reconnect_counter()
        self.__add_on_channel_close_callback()
        self.__add_on_return_callback()
        self.__make_channel_confirm_delivery()
        self.__make_ready_for_publishing()

    '''
    Once we succeeded in building a connection, we reset the
    reconnection counter, so after a connection was interrupted,
    we can do the same number of reconnection attempts again.
    This is called when a connection AND channel was successfully
    build, i.e. in the on_channel_open workflow.
    '''
    def __reset_reconnect_counter(self):
        logdebug(LOGGER, 'Resetting reconnection counter, because a channel was successfully opened.')
        self.__backup_reconnect_counter = self.__reconnect_counter # we may need to undo this later...
        self.__reconnect_counter = 0

    '''
    Sometimes we assume we were successful in connecting and set the
    reconnection counter to zero.
    But the connection was not successful, so we reconnect, and as the
    reconnection counter was reset, we do so infinitely.

    This occurs if we did succeed in opening a channel, but the
    host we connected to lacks the required exchange. This is only
    noticed when we try to send the first message. Then, the channel
    is closed. So we call this in the on_channel_closed, but only
    in specific situations.
    '''
    def __undo_resetting_reconnect_counter(self):
        logdebug(LOGGER, 'Undo resetting reconnection counter, because the channel that was opened did not actually function.')
        self.__reconnect_counter = self.__backup_reconnect_counter

    def __make_channel_confirm_delivery(self):
        logtrace(LOGGER, 'Set confirm delivery... (Issue Confirm.Select RPC command)')
        self.thread._channel.confirm_delivery(ack_nack_callback=self.confirmer.on_delivery_confirmation)
        logdebug(LOGGER, 'Set confirm delivery... done.')
 
    def __make_ready_for_publishing(self):
        logdebug(LOGGER, '(Re)connection established, making ready for publication...')

        # Check for unexpected errors:
        if self.thread._channel is None:
            logerror(LOGGER, 'Channel is None after connecting to server. This should not happen.')
            self.statemachine.set_to_permanently_unavailable()
        if self.thread._connection is None:
            logerror(LOGGER, 'Connection is None after connecting to server. This should not happen.')
            self.statemachine.set_to_permanently_unavailable()

        # Normally, it should already be waiting to be available:
        if self.statemachine.is_WAITING_TO_BE_AVAILABLE():
            logdebug(LOGGER, 'Setup is finished. Publishing may start.')
            logtrace(LOGGER, 'Publishing will use channel no. %s!', self.thread._channel.channel_number)
            self.statemachine.set_to_available()
            self.__check_for_already_arrived_messages_and_publish_them()

        # It was asked to close in the meantime (but might be able to publish the last messages):
        elif self.statemachine.is_AVAILABLE_BUT_WANTS_TO_STOP():
            logdebug(LOGGER, 'Setup is finished, but the module was already asked to be closed in the meantime.')
            self.__check_for_already_arrived_messages_and_publish_them()

        # It was force-closed in the meantime:
        elif self.statemachine.is_PERMANENTLY_UNAVAILABLE() or self.statemachine.is_FORCE_FINISHED():
            if self.statemachine.get_detail_closed_by_publisher():
                logdebug(LOGGER, 'Setup is finished now, but the module was already force-closed in the meantime.')
                self.shutter.safety_finish('closed before connection was ready. reclosing.')
            elif self.statemachine.detail_could_not_connect:
                logerror(LOGGER, 'This is not supposed to happen. If the connection failed, this part of the code should not be reached.')
            else:
                logerror(LOGGER, 'This is not supposed to happen. An unknown event set this module to be unavailable. When was this set to unavailable?')
        else:
            logdebug(LOGGER, 'Unexpected state.')

    def __check_for_already_arrived_messages_and_publish_them(self):
        logdebug(LOGGER, 'Checking if messages have arrived in the meantime...')
        num = self.thread.get_num_unpublished()
        if num > 0:
            loginfo(LOGGER, 'Ready to publish messages to RabbitMQ. %s messages are already waiting to be published.', num)
            for i in range(int(num*1.1)):
                self.thread.add_event_publish_message()
        else:
            loginfo(LOGGER, 'Ready to publish messages to RabbitMQ.')
            logdebug(LOGGER, 'Ready to publish messages to RabbitMQ. No messages waiting yet.')
        

    ########################
    ### Connection error ###
    ########################

    '''
    If the connection to RabbitMQ failed, there is various
    things that may happen:
    (1) If there is other RabbitMQ urls, it will try to connect 
        to one of these.
    (2) If there is no other URLs, it will try to reconnect to this
        one after a short waiting time.
    (3) If the maximum number of reconnection tries is reached, it
        gives up.
    '''
    def on_connection_error(self, connection, msg):

        oldhost = self.__get_whole_host_name()
        time_passed = datetime.datetime.now() - self.__start_connect_time
        time_passed_seconds = time_passed.total_seconds()
        logerror(LOGGER, 'Could not connect to %s: "%s" (connection failure after %s seconds)', oldhost, msg, time_passed_seconds)

        self.__store_connection_error_info(msg, oldhost)

        # If there was a force-finish, we do not reconnect.
        if self.statemachine.is_FORCE_FINISHED():
            errormsg = 'Permanently failed to connect to RabbitMQ.'
            if self.statemachine.detail_asked_to_gently_close_by_publisher:
                errormsg += ' Tried all hosts until was force-closed by user.'
            elif self.statemachine.detail_asked_to_force_close_by_publisher:
                errormsg += ' Tried all hosts until a user close-down forced us to give up (e.g. the maximum waiting time was reached).'
            errormsg += ' Giving up. No PID requests will be sent.'
            self.__give_up_reconnecting_and_raise_exception(errormsg)
        
        # If there is alternative URLs, try one of them:
        if self.__node_manager.has_more_urls():
            logdebug(LOGGER, 'Connection failure: %s fallback URLs left to try.', self.__node_manager.get_num_left_urls())
            self.__node_manager.set_next_host()
            newhost = self.__get_whole_host_name()
            loginfo(LOGGER, 'Connection failure: Trying to connect (now) to %s.', newhost)
            reopen_seconds = 0
            self.__wait_and_trigger_reconnection(connection, reopen_seconds)


        # If there is no URLs, reset the node manager to
        # start at the first nodes again...
        else:
            self.__reconnect_counter += 1;
            if self.__reconnect_counter <= self.__max_reconnection_tries:
                reopen_seconds = self.__wait_seconds_before_reconnect
                logdebug(LOGGER, 'Connection failure: Failed connecting to all hosts. Waiting %s seconds and starting over.', reopen_seconds)
                self.__node_manager.reset_nodes()
                newhost = self.__node_manager.get_connection_parameters().host
                loginfo(LOGGER, 'Connection failure: Trying to connect (in %s seconds) to %s.', reopen_seconds, newhost)
                self.__wait_and_trigger_reconnection(connection, reopen_seconds)

            # Give up after so many tries...
            else:
                errormsg = ('Permanently failed to connect to RabbitMQ. Tried all hosts %s times. Giving up. No PID requests will be sent.' % (self.__max_reconnection_tries+1))
                self.__give_up_reconnecting_and_raise_exception(errormsg)

    def __give_up_reconnecting_and_raise_exception(self, error_message):
        self.statemachine.set_to_permanently_unavailable()
        self.statemachine.detail_could_not_connect = True
        problem_message = self.__connection_errors_to_string()
        logerror(LOGGER, error_message)
        logdebug(LOGGER, problem_message)
        self.__make_permanently_closed_by_error(None, self.thread.ERROR_TEXT_CONNECTION_PERMANENT_ERROR) # Stops ioloop, so thread may stop!
        if not (hasattr(defaults, 'IS_TEST_RUN') and defaults.IS_TEST_RUN==True):
            raise PIDServerException(error_message+'\nProblems:\n'+problem_message)
        else:
            msg = 'PIDServerException would have been raised in real life.'
            logerror(LOGGER, msg)
            
    def __store_connection_error_info(self, errorname, host):
        errorname = str(errorname)
        if host not in self.__connection_errors:
            self.__connection_errors[host] = {}
        if errorname not in self.__connection_errors[host]:
            self.__connection_errors[host][errorname] = 1
        else: 
            self.__connection_errors[host][errorname] += 1

    def __connection_errors_to_string(self, sep='\n'):
        separate_messages_per_host = []
 
        # For each host:
        for host in self.__connection_errors:
            all_errors_for_one_host = []
 
            # For each error in this host:
            for errortype in self.__connection_errors[host]:
                num = self.__connection_errors[host][errortype]
                message_for_one_error_type = ('%ix "%s"' % (num, errortype))
                all_errors_for_one_host.append(message_for_one_error_type)
 
            concat_errors = ', '.join(all_errors_for_one_host)
            message_for_one_host = 'Server "%s": %s' % (host, concat_errors)
            separate_messages_per_host.append(message_for_one_host)
 
        return sep.join(separate_messages_per_host)

    def __get_whole_host_name(self):
        params = self.__node_manager.get_connection_parameters()
        name = params.host
        vhost = params.virtual_host
        if vhost =='/':
            pass
        elif vhost.startswith('/'):
            name += vhost
        else:
            name += '/'+vhost
        name += ':'+str(params.port)
        return name


    #############################
    ### React to channel and  ###
    ### connection close      ###
    #############################

    ''' This tells RabbitMQ what to do if it receives 
    a message it cannot accept, e.g. if it cannot
    route it. '''
    def __add_on_return_callback(self):
        self.thread._channel.add_on_return_callback(self.returnhandler.on_message_not_accepted)

    '''
    This tells RabbitMQ what to do if the channel
    was closed.

    Note: Every connection close includes a channel close.
    However, as far as I know, this callback is only
    called if the channel is closed without the underlying
    connection being closed. I am not 100 percent sure though.
    '''
    def __add_on_channel_close_callback(self):
        self.thread._channel.add_on_close_callback(self.on_channel_closed)

    '''
    Callback, called by RabbitMQ.
    "on_channel_closed" can be called in three situations:

    (1) The user asked to close the connection.
        In this case, we want to clean up everything and leave it closed.

    (2) The connection was closed because we tried to publish to a non-
        existent exchange.
        In this case, the connection is still open, and we want to reopen
        a new channel and publish to a different exchange.
        We also want to republish the ones that had failed.

    (3) There was some problem that closed the connection, which causes
        the channel to close.
        In this case, we want to reopen a connection.

    '''
    def on_channel_closed(self, connection, exception):

        # From the docs: The exception will either be an instance of
        # exceptions.ConnectionClosed if a fully-open connection was closed
        # by user or broker or exception of another type (...)
        if isinstance(exception, pika.exceptions.ChannelClosed):
            reply_code = exception.reply_code
            reply_text = exception.reply_text
        else:
            # TODO Not sure when this might happen, could not reproduce.
            reply_code = -1
            reply_text = str(exception)

        logdebug(LOGGER, 'Channel was closed: %s (code %s)', reply_text, reply_code)

        # Channel closed because user wants to close:
        if self.statemachine.is_PERMANENTLY_UNAVAILABLE() or self.statemachine.is_FORCE_FINISHED():
            if self.statemachine.get_detail_closed_by_publisher():
                logdebug(LOGGER,'Channel close event due to close command by user. This is expected.')

        # Channel closed because even fallback exchange did not exist:
        elif reply_code == 404 and "NOT_FOUND - no exchange 'FALLBACK'" in reply_text:
            logerror(LOGGER,'Channel closed because FALLBACK exchange does not exist. Need to close connection to trigger all the necessary close down steps.')
            self.__undo_resetting_reconnect_counter()
            self.thread.reset_exchange_name() # So next host is tried with normal exchange
            self.thread._connection.close() # This will reconnect!
            # TODO: Put a different reply_code and text, so we won't treat this as a Normal Shutdown!

        # Channel closed because exchange did not exist:
        elif reply_code == 404:
            logdebug(LOGGER, 'Channel closed because the exchange "%s" did not exist.', self.__node_manager.get_exchange_name())
            self.__use_different_exchange_and_reopen_channel()

        # Other unexpected channel close:
        else:
            logerror(LOGGER,'Unexpected channel shutdown. Need to close connection to trigger all the necessary close down steps.')
            self.__undo_resetting_reconnect_counter()
            self.thread._connection.close() # This will reconnect!
            # TODO: Put a different reply_code and text, so we won't treat this as a Normal Shutdown!

    '''
    An attempt to publish to a nonexistent exchange will close
    the channel. In this case, we use a different exchange name
    and reopen the channel. The underlying connection was kept
    open.
    '''
    def __use_different_exchange_and_reopen_channel(self):

        # Set to waiting to be available, so that incoming
        # messages are stored:
        self.statemachine.set_to_waiting_to_be_available()

        # New exchange name
        logdebug(LOGGER, 'Setting exchange name to fallback exchange "%s"', self.__fallback_exchange_name)
        self.thread.change_exchange_name(self.__fallback_exchange_name)

        # If this happened while sending message to the wrong exchange, we
        # have to trigger their resending...
        self.__prepare_channel_reopen('Channel reopen')

        # Reopen channel
        logdebug(LOGGER, 'Reopening channel...')
        self.statemachine.set_to_waiting_to_be_available()
        self.__please_open_rabbit_channel()

    '''
    Callback, called by RabbitMQ.
    "on_connection_closed" can be called in two situations:

    (1) The user asked to close the connection.
        In this case, we want to clean up everything and leave it closed.

    (2) There was some other problem that closed the connection.

    '''
    def on_connection_closed(self, connection, exception):

        # From the docs: The exception will either be an instance of
        # exceptions.ConnectionClosed if a fully-open connection was closed
        # by user or broker or exception of another type (...)
        if isinstance(exception, pika.exceptions.ConnectionClosed):
            reply_code = exception.reply_code
            reply_text = exception.reply_text
        else:
            # TODO Not sure when this might happen, could not reproduce.
            reply_code = -1 
            reply_text = str(exception)

        loginfo(LOGGER, 'Connection to RabbitMQ was closed. Reason: %s.', reply_text)
        self.thread._channel = None
        if self.__was_user_shutdown(reply_code, reply_text):
            loginfo(LOGGER, 'Connection to %s closed.', self.__node_manager.get_connection_parameters().host)
            self.make_permanently_closed_by_user()
        elif self.__was_permanent_error(reply_code, reply_text):
            loginfo(LOGGER, 'Connection to %s closed.', self.__node_manager.get_connection_parameters().host)
            self.__make_permanently_closed_by_error(connection, reply_text)
        else:
            # This reconnects to next host:
            self.on_connection_error(connection, reply_text)

    def __was_permanent_error(self, reply_code, reply_text):
        if self.thread.ERROR_TEXT_CONNECTION_PERMANENT_ERROR in reply_text:
            return True
        return False

    def __was_user_shutdown(self, reply_code, reply_text):
        if self.__was_forced_user_shutdown(reply_code, reply_text):
            return True
        elif self.__was_gentle_user_shutdown(reply_code, reply_text):
            return True
        return False

    def __was_forced_user_shutdown(self, reply_code, reply_text):
        if (reply_code==self.thread.ERROR_CODE_CONNECTION_CLOSED_BY_USER and
            self.thread.ERROR_TEXT_CONNECTION_FORCE_CLOSED in reply_text):
            return True
        return False

    def __was_gentle_user_shutdown(self, reply_code, reply_text):
        if (reply_code==self.thread.ERROR_CODE_CONNECTION_CLOSED_BY_USER and
            self.thread.ERROR_TEXT_CONNECTION_NORMAL_SHUTDOWN in reply_text):
            return True
        return False

    ''' Called by thread, by shutter module.'''
    def make_permanently_closed_by_user(self):
        # This changes the state of the state machine!
        # This needs to be called from the shutter module
        # in case there is a force_finish while the connection
        # is already closed (as the callback on_connection_closed
        # is not called then).
        self.statemachine.set_to_permanently_unavailable()
        logtrace(LOGGER, 'Stop waiting for events due to user interrupt!')
        logtrace(LOGGER, 'Permanent close: Stopping ioloop of connection %s...', self.thread._connection)
        self.thread._connection.ioloop.stop()
        loginfo(LOGGER, 'Stopped listening for RabbitMQ events (%s).', get_now_utc_as_formatted_string())
        logdebug(LOGGER, 'Connection to messaging service closed by user. Will not reopen.')

    def __make_permanently_closed_by_error(self, connection, reply_text):
        # This changes the state of the state machine!
        # This needs to be called if there is a permanent
        # error and we don't want the library to reonnect,
        # and we also don't want to pretend it was closed
        # by the user.
        # This is really rarely needed. 
        self.statemachine.set_to_permanently_unavailable()
        logtrace(LOGGER, 'Stop waiting for events due to permanent error!')

        # In case the main thread was waiting for any synchronization event.
        self.thread.unblock_events()

        # Close ioloop, which blocks the thread.
        logdebug(LOGGER, 'Permanent close: Stopping ioloop of connection %s...', self.thread._connection)
        self.thread._connection.ioloop.stop()
        loginfo(LOGGER, 'Stopped listening for RabbitMQ events (%s).', get_now_utc_as_formatted_string())
        logdebug(LOGGER, 'Connection to messaging service closed because of error. Will not reopen. Reason: %s', reply_text)

    '''
    This triggers a reconnection to whatever host is stored in
    self.__node_manager.get_connection_parameters().host at the moment of reconnection.

    If it is called to reconnect to the same host, it is better
    to wait some seconds.

    If it is used to connect to the next host, there is no point
    in waiting.
    '''
    def __wait_and_trigger_reconnection(self, connection, wait_seconds):
        if self.statemachine.is_FORCE_FINISHED():
            errormsg = 'Permanently failed to connect to RabbitMQ. Tried all hosts until received a force-finish. Giving up. No PID requests will be sent.'
            self.__give_up_reconnecting_and_raise_exception(errormsg)
        else:
            self.statemachine.set_to_waiting_to_be_available()
            loginfo(LOGGER, 'Trying to reconnect to RabbitMQ in %s seconds.', wait_seconds)
            connection.ioloop.call_later(wait_seconds, self.reconnect)
            logtrace(LOGGER, 'Reconnect event added to connection %s (not to %s)', connection, self.thread._connection)

    ###########################
    ### Reconnect after     ###
    ### unexpected shutdown ###
    ###########################

    '''
    Reconnecting creates a completely new connection.
    If we reconnect, we need to reset message number,
    delivery tag etc.

    We need to prepare to republish the yet-unconfirmed
    messages.

    Then we need to stop the old connection's ioloop.
    The reconnection will create a new connection object
    and this will have its own ioloop.

    '''
    def reconnect(self):
        logdebug(LOGGER, 'Reconnecting...')

        # We need to reset delivery tags, unconfirmed messages,
        # republish the unconfirmed, ...
        self.__prepare_channel_reopen('Reconnect')
        
        # This is the old connection ioloop instance, stop its ioloop
        logdebug(LOGGER, 'Reconnect: Stopping ioloop of connection %s...', self.thread._connection)
        self.thread._connection.ioloop.stop()
        # Note: All events still waiting on the ioloop are lost.
        # Messages are kept track of in the Queue.Queue or in the confirmer
        # module. Closing events are kept track on in shutter module.

        # Now we trigger the actual reconnection, which
        # works just like the first connection to RabbitMQ.
        self.first_connection()

    '''
    This is called during reconnection and during channel reopen.
    Both implies that a new channel is opened.
    '''
    def __prepare_channel_reopen(self, operation_string):
        # We need to reset the message number, as
        # it works by channel:
        logdebug(LOGGER, operation_string+': Resetting delivery number (for publishing messages).')
        self.thread.reset_delivery_number()

        # Furthermore, as we'd like to re-publish messages
        # that had not been confirmed yet, we remove them
        # from the stack of unconfirmed messages, and put them
        # back to the stack of unpublished messages.
        logdebug(LOGGER, operation_string+': Sending all messages that have not been confirmed yet...')
        self.__prepare_republication_of_unconfirmed()

        # Reset the unconfirmed delivery tags, as they also work by channel:
        logdebug(LOGGER, operation_string+': Resetting delivery tags (for confirming messages).')
        self.thread.reset_unconfirmed_messages_and_delivery_tags()
        
    def __prepare_republication_of_unconfirmed(self):
        # Get all unconfirmed messages - we won't be able to receive their confirms anymore:
        # IMPORTANT: This has to happen before we reset the delivery_tags of the confirmer
        # module, as this deletes the collection of unconfirmed messages.
        rescued_messages = self.thread.get_unconfirmed_messages_as_list_copy_during_lifetime()
        if len(rescued_messages)>0:
            logdebug(LOGGER, '%s unconfirmed messages were saved and are sent now.', len(rescued_messages))
            self.thread.send_many_messages(rescued_messages)
            # Note: The actual publish of these messages to rabbit
            # happens when the connection is there again, so no wrong delivery
            # tags etc. are created by this line!






import unittest
import logging
import pika
import esgfpid.rabbit
import tests.globalvar

LOGGER = logging.getLogger(__name__)
LOGGER.addHandler(logging.NullHandler())

# Test resources:
import resources.TESTVALUES as TESTHELPERS

# Some tests rely on open nodes
import globalvar
if globalvar.RABBIT_OPEN_NOT_ALLOWED:
    print('Skipping tests that need open RabbitMQ nodes in module "%s".' % __name__)


'''
Unit tests for esgfpid.rabbit.nodemanager.

This module does not need any other module to
function, so we don't need to use or to mock any
other objects. 
'''
class NodemanagerTestCase(unittest.TestCase):

    def setUp(self):
        LOGGER.info('######## Next test (%s) ##########', __name__)

    def tearDown(self):
        LOGGER.info('#############################')

    #
    # Test the constructor.
    #

    '''
    Simple test whether the constructor works.
    It takes no args at all.
    '''
    def test_constructor(self):

        # Run code to be tested:
        mynodemanager = esgfpid.rabbit.nodemanager.NodeManager()

        # Check result
        self.assertIsInstance(mynodemanager, esgfpid.rabbit.nodemanager.NodeManager)

    '''
    Test if I can pass a trusted node with no priority.
    '''
    def test_add_one_trusted_node_no_prio_ok(self):

        # Test variables:
        mynodemanager = esgfpid.rabbit.nodemanager.NodeManager()

        # Run code to be tested:
        args = TESTHELPERS.get_args_for_nodemanager()
        mynodemanager.add_trusted_node(**args)

        # Check whether the correct number of nodes is there:
        self.assertEqual(mynodemanager.get_num_left_open(),0)
        self.assertEqual(mynodemanager.get_num_left_trusted(),1)
        self.assertEqual(mynodemanager.get_num_left_urls(),1)

        # Check whether the correct priorities are set:
        mynodemanager.set_next_host()
        node = mynodemanager._NodeManager__current_node
        self.assertEqual(node['priority'], 'zzz_default')
        self.assertFalse(node['is_open'])
        self.assertIsInstance(node['credentials'], pika.PlainCredentials)
        self.assertIsInstance(node['params'], pika.ConnectionParameters)

    '''
    Test if I can pass a trusted node with priority=None.
    '''
    def test_add_one_trusted_node_prio_none_ok(self):

        # Test variables:
        mynodemanager = esgfpid.rabbit.nodemanager.NodeManager()
        args = TESTHELPERS.get_args_for_nodemanager(priority=None, vhost='foo', port=22, ssl_enabled=True)

        # Run code to be tested:
        mynodemanager.add_trusted_node(**args)

        # Check whether the correct number of nodes is there:
        self.assertEqual(mynodemanager.get_num_left_open(),0)
        self.assertEqual(mynodemanager.get_num_left_trusted(),1)
        self.assertEqual(mynodemanager.get_num_left_urls(),1)

        # Check whether the correct priorities are set:
        mynodemanager.set_next_host()
        node = mynodemanager._NodeManager__current_node
        self.assertEqual(node['priority'], 'zzz_default')
        self.assertFalse(node['is_open'])
        self.assertIsInstance(node['credentials'], pika.PlainCredentials)
        self.assertIsInstance(node['params'], pika.ConnectionParameters)
        self.assertEqual(node['vhost'], 'foo')
        self.assertEqual(node['port'], 22)
        self.assertEqual(node['ssl_enabled'], True)

    '''
    Test if I can pass several trusted nodes with no priority.
    '''
    def test_add_trusted_nodes_no_prio_ok(self):

        # Test variables:
        # Make test node manager:
        mynodemanager = esgfpid.rabbit.nodemanager.NodeManager()
        args = TESTHELPERS.get_args_for_nodemanager()

        # Run code to be tested:
        mynodemanager.add_trusted_node(**args)
        mynodemanager.add_trusted_node(**args)

        # Check whether the correct number of nodes is there:
        self.assertEqual(mynodemanager.get_num_left_open(),0)
        self.assertEqual(mynodemanager.get_num_left_trusted(),2)
        self.assertEqual(mynodemanager.get_num_left_urls(),2)

    '''
    Test if I can pass several trusted nodes with priority.
    '''
    def test_add_trusted_nodes_with_prio_ok(self):

        # Test variables:
        # Make test node manager, and node-dictionaries with
        # various priorities:
        mynodemanager = esgfpid.rabbit.nodemanager.NodeManager()
        args1 = TESTHELPERS.get_args_for_nodemanager(priority=3)
        args2 = TESTHELPERS.get_args_for_nodemanager(priority=1)
        # Run code to be tested:
        mynodemanager.add_trusted_node(**args1)
        mynodemanager.add_trusted_node(**args2)

        # Check whether the correct number of nodes is there:
        self.assertEqual(mynodemanager.get_num_left_open(),0)
        self.assertEqual(mynodemanager.get_num_left_trusted(),2)
        self.assertEqual(mynodemanager.get_num_left_urls(),2)
        # Check whether the correct priorities are set:
        mynodemanager.set_next_host()
        node = mynodemanager._NodeManager__current_node
        self.assertEqual(node['priority'], '1')
        mynodemanager.set_next_host()
        node = mynodemanager._NodeManager__current_node
        self.assertEqual(node['priority'], '3')

    '''
    Test exception if I miss info
    '''
    def test_add_nodes_missing_info(self):

        # Test variables:
        # Make test node manager:
        mynodemanager = esgfpid.rabbit.nodemanager.NodeManager()

        # Run code to be tested and check exception:
        with self.assertRaises(esgfpid.exceptions.ArgumentError):
            mynodemanager.add_trusted_node(username='foo', exchange_name='bar')

    '''
    Test exception if I miss info, and check if password is printed
    (it should not!)
    '''
    def test_add_nodes_missing_info_2(self):

        # Test variables:
        # Make test node manager:
        mynodemanager = esgfpid.rabbit.nodemanager.NodeManager()

        # Run code to be tested and check exception:
        with self.assertRaises(esgfpid.exceptions.ArgumentError) as context_manager:
            mynodemanager.add_trusted_node(username='foo', exchange_name='bar',
                password='einekleinenachtmusik')

        # Check exception content:
        ex = context_manager.exception
        self.assertTrue('einekleinenachtmusik' not in ex.msg, ex)

    '''
    Test exception if I miss info
    '''
    @unittest.skipIf(globalvar.RABBIT_OPEN_NOT_ALLOWED, '(this test uses open rabbit nodes)')
    def test_add_nodes_missing_info_open(self):

        # Test variables:
        # Make test node manager:
        mynodemanager = esgfpid.rabbit.nodemanager.NodeManager()

        # Run code to be tested and check exception:
        with self.assertRaises(esgfpid.exceptions.ArgumentError):
            mynodemanager.add_trusted_node(username='foo', exchange_name='bar')
        with self.assertRaises(esgfpid.exceptions.ArgumentError):
            mynodemanager.add_open_node(username='foo', exchange_name='bar')


    '''
    Test the algorithm that picks the next host, based on
    passed priorities.
    '''
    def test_priority_algorithm_trusted_only(self):

        # Test variables:
        mynodemanager = esgfpid.rabbit.nodemanager.NodeManager()
        args1 = TESTHELPERS.get_args_for_nodemanager(host='foo1', priority=1)
        args2 = TESTHELPERS.get_args_for_nodemanager(host='foo2', priority=3)
        args3 = TESTHELPERS.get_args_for_nodemanager(host='foo3', priority=3)
        args4 = TESTHELPERS.get_args_for_nodemanager(host='foo4')
        # Add the nodes, in no particular order:
        mynodemanager.add_trusted_node(**args2)
        mynodemanager.add_trusted_node(**args4)
        mynodemanager.add_trusted_node(**args3)
        mynodemanager.add_trusted_node(**args1)

        # Run code to be tested:
        # Set one host after the other and check whether the
        # correct host was picked, based on the passed priority.
        # The first one should be foo1
        mynodemanager.set_next_host()
        node = mynodemanager._NodeManager__current_node
        self.assertEqual(node['host'], 'foo1')
        self.assertFalse(node['is_open'])
        # The second one should be foo2 OR foo3 (same prio)
        mynodemanager.set_next_host()
        node = mynodemanager._NodeManager__current_node
        self.assertTrue('foo2' in node['host'] or 'foo3' in node['host'])
        self.assertFalse(node['is_open'])
        # The third one should be foo2 OR foo3 (same prio)
        mynodemanager.set_next_host()
        node = mynodemanager._NodeManager__current_node
        self.assertTrue('foo2' in node['host'] or 'foo3' in node['host'])
        self.assertFalse(node['is_open'])
        # The fourth one should be foo4
        mynodemanager.set_next_host()
        node = mynodemanager._NodeManager__current_node
        self.assertEqual(node['host'], 'foo4')
        self.assertFalse(node['is_open'])


    '''
    Test the algorithm that picks the next host, based on
    passed priorities.
    '''
    @unittest.skipIf(globalvar.RABBIT_OPEN_NOT_ALLOWED, '(this test uses open rabbit nodes)')
    def test_priority_algorithm_open(self):

        # Test variables:
        mynodemanager = esgfpid.rabbit.nodemanager.NodeManager()
        args1 = TESTHELPERS.get_args_for_nodemanager(host='foo1', priority=1)
        args2 = TESTHELPERS.get_args_for_nodemanager(host='foo2', priority=3)
        args3 = TESTHELPERS.get_args_for_nodemanager(host='foo3', priority=3)
        args4 = TESTHELPERS.get_args_for_nodemanager(host='foo4')
        args1o = TESTHELPERS.get_args_for_nodemanager(host='foo1o', priority=1)
        args2o = TESTHELPERS.get_args_for_nodemanager(host='foo2o', priority=3)
        args3o = TESTHELPERS.get_args_for_nodemanager(host='foo3o', priority=3)
        args4o = TESTHELPERS.get_args_for_nodemanager(host='foo4o')
        # Add the nodes, in no particular order:
        mynodemanager.add_open_node(**args2o)
        mynodemanager.add_open_node(**args4o)
        mynodemanager.add_open_node(**args1o)
        mynodemanager.add_open_node(**args3o)
        mynodemanager.add_trusted_node(**args2)
        mynodemanager.add_trusted_node(**args4)
        mynodemanager.add_trusted_node(**args3)
        mynodemanager.add_trusted_node(**args1)

        # Run code to be tested:
        # Set one host after the other and check whether the
        # correct host was picked, based on the passed priority.
        # The first one should be foo1
        mynodemanager.set_next_host()
        node = mynodemanager._NodeManager__current_node
        self.assertEqual(node['host'], 'foo1')
        self.assertFalse(node['is_open'])
        # The second one should be foo2 OR foo3 (same prio)
        mynodemanager.set_next_host()
        node = mynodemanager._NodeManager__current_node
        self.assertTrue('foo2' in node['host'] or 'foo3' in node['host'])
        self.assertFalse(node['is_open'])
        # The third one should be foo2 OR foo3 (same prio)
        mynodemanager.set_next_host()
        node = mynodemanager._NodeManager__current_node
        self.assertTrue('foo2' in node['host'] or 'foo3' in node['host'])
        self.assertFalse(node['is_open'])
        # The fourth one should be foo4
        mynodemanager.set_next_host()
        node = mynodemanager._NodeManager__current_node
        self.assertEqual(node['host'], 'foo4')
        self.assertFalse(node['is_open'])
        # The fifth one should be foo1o (the first open)
        mynodemanager.set_next_host()
        node = mynodemanager._NodeManager__current_node
        self.assertEqual(node['host'], 'foo1o')
        self.assertTrue(node['is_open'])
        # The sixth one should be foo2o OR foo3o (same prio)
        mynodemanager.set_next_host()
        node = mynodemanager._NodeManager__current_node
        self.assertTrue('foo2o' in node['host'] or 'foo3o' in node['host'])
        self.assertTrue(node['is_open'])
        # The seventh one should be foo2 OR foo3 (same prio)
        mynodemanager.set_next_host()
        node = mynodemanager._NodeManager__current_node
        self.assertTrue('foo2o' in node['host'] or 'foo3o' in node['host'])
        self.assertTrue(node['is_open'])
        # The eigth one should be foo4o
        mynodemanager.set_next_host()
        node = mynodemanager._NodeManager__current_node
        self.assertEqual(node['host'], 'foo4o')
        self.assertTrue(node['is_open'])

    '''
    Test getter for exchange name.
    '''
    def test_get_exchange_name(self):

        # Make node manager:
        mynodemanager = esgfpid.rabbit.nodemanager.NodeManager()
        # Add the nodes, in no particular order:
        args1 = TESTHELPERS.get_args_for_nodemanager(exchange_name='foo2', priority=6)
        args2 = TESTHELPERS.get_args_for_nodemanager(exchange_name='foo3', priority=1)
        args3 = TESTHELPERS.get_args_for_nodemanager(exchange_name='foo1', priority=2)
        mynodemanager.add_trusted_node(**args1)
        mynodemanager.add_trusted_node(**args2)
        mynodemanager.add_trusted_node(**args3)

        # Run code to be tested:
        # Set one host after the other and check if the correct
        # exchange name is returned.
        mynodemanager.set_next_host()
        self.assertEqual('foo3', mynodemanager.get_exchange_name())
        mynodemanager.set_next_host()
        self.assertEqual('foo1', mynodemanager.get_exchange_name())
        mynodemanager.set_next_host()
        self.assertEqual('foo2', mynodemanager.get_exchange_name())

    '''
    Test getter for exchange name.
    '''
    @unittest.skipIf(globalvar.RABBIT_OPEN_NOT_ALLOWED, '(this test uses open rabbit nodes)')
    def test_get_exchange_name_open(self):

        # Make node manager:
        mynodemanager = esgfpid.rabbit.nodemanager.NodeManager()
        # Add the nodes, in no particular order:
        args1 = TESTHELPERS.get_args_for_nodemanager(exchange_name='foo2', priority=6)
        args2 = TESTHELPERS.get_args_for_nodemanager(exchange_name='foo3', priority=1)
        args3 = TESTHELPERS.get_args_for_nodemanager(exchange_name='foo1', priority=2)
        mynodemanager.add_trusted_node(**args1)
        mynodemanager.add_open_node(**args2)
        mynodemanager.add_trusted_node(**args3)

        # Run code to be tested:
        # Set one host after the other and check if the correct
        # exchange name is returned.
        mynodemanager.set_next_host()
        self.assertEqual('foo1', mynodemanager.get_exchange_name())
        mynodemanager.set_next_host()
        self.assertEqual('foo2', mynodemanager.get_exchange_name())
        mynodemanager.set_next_host()
        self.assertEqual('foo3', mynodemanager.get_exchange_name())

    '''
    Test behaviour of set_next_host() if none is left.
    '''
    def test_setting_nodes_if_none_left(self):

        # Make a node manager with several nodes, and remove them
        # one by one.
        mynodemanager = esgfpid.rabbit.nodemanager.NodeManager()
        args1 = TESTHELPERS.get_args_for_nodemanager(exchange_name='foo', priority=1)
        args2 = TESTHELPERS.get_args_for_nodemanager(exchange_name='bar', priority=2)
        mynodemanager.add_trusted_node(**args1)
        mynodemanager.add_trusted_node(**args2)

        # Run code to be tested and check results:
        # Set the first host, with exchange=foo
        mynodemanager.set_next_host()
        self.assertEqual(mynodemanager.get_exchange_name(), 'foo')
        # Set the second host, with exchange=bar
        mynodemanager.set_next_host()
        self.assertEqual(mynodemanager.get_exchange_name(), 'bar')
        # If we continue trying to set a next one, it stays "bar",
        # as no other is left!
        mynodemanager.set_next_host()
        self.assertEqual(mynodemanager.get_exchange_name(), 'bar')
        mynodemanager.set_next_host()
        self.assertEqual(mynodemanager.get_exchange_name(), 'bar')
        mynodemanager.set_next_host()
        self.assertEqual(mynodemanager.get_exchange_name(), 'bar')

    '''
    Test whether setting and resetting nodes works.
    '''
    def test_setting_and_resetting_nodes_trusted_only(self):

        # Make a node manager with several nodes, and remove them
        # one by one.
        mynodemanager = esgfpid.rabbit.nodemanager.NodeManager()
        mynodemanager.add_trusted_node(**TESTHELPERS.get_args_for_nodemanager())
        mynodemanager.add_trusted_node(**TESTHELPERS.get_args_for_nodemanager())
        mynodemanager.add_trusted_node(**TESTHELPERS.get_args_for_nodemanager())
        mynodemanager.add_trusted_node(**TESTHELPERS.get_args_for_nodemanager())
        mynodemanager.add_trusted_node(**TESTHELPERS.get_args_for_nodemanager())

        # Run code to be tested:
        # First, set one after the other, and check if the
        # number of URLs left is correct.
        # When all are used, reset the hosts and check if
        # the resetting worked fine.

        # We added five hosts, 3 trusted and 2 untrusted:
        self.assertEqual(mynodemanager.get_num_left_urls(),5)
        self.assertEqual(mynodemanager.get_num_left_trusted(),5)
        self.assertEqual(mynodemanager.get_num_left_open(),0)
        self.assertTrue(mynodemanager.has_more_urls())
        # Set the first host (1/5), 4 are left. (First, the trusted ones are selected):
        mynodemanager.set_next_host()
        self.assertEqual(mynodemanager.get_num_left_urls(),4)
        self.assertEqual(mynodemanager.get_num_left_trusted(),4)
        self.assertEqual(mynodemanager.get_num_left_open(),0)
        self.assertTrue(mynodemanager.has_more_urls())
        # Set the second host (2/5), 3 are left.
        mynodemanager.set_next_host()
        self.assertEqual(mynodemanager.get_num_left_urls(),3)
        self.assertEqual(mynodemanager.get_num_left_trusted(),3)
        self.assertEqual(mynodemanager.get_num_left_open(),0)
        # Set the third host (3/5), 2 are left (the two untrusted ones):
        mynodemanager.set_next_host()
        self.assertEqual(mynodemanager.get_num_left_urls(),2)
        self.assertEqual(mynodemanager.get_num_left_trusted(),2)
        self.assertEqual(mynodemanager.get_num_left_open(),0)
        self.assertTrue(mynodemanager.has_more_urls())
        # Set the fourth host (4/5), 1 is left:
        mynodemanager.set_next_host()
        self.assertEqual(mynodemanager.get_num_left_urls(),1)
        self.assertEqual(mynodemanager.get_num_left_trusted(),1)
        self.assertEqual(mynodemanager.get_num_left_open(),0)
        self.assertTrue(mynodemanager.has_more_urls())
        # Set the last host (5/5), so 0 are left:
        mynodemanager.set_next_host()
        self.assertEqual(mynodemanager.get_num_left_urls(),0)
        self.assertEqual(mynodemanager.get_num_left_trusted(),0)
        self.assertEqual(mynodemanager.get_num_left_open(),0)
        self.assertFalse(mynodemanager.has_more_urls())
        # Reset all nodes.
        # This includes the first call to set_next_host(), so it 
        # sets the first of five, so 4 are left:
        mynodemanager.reset_nodes()
        self.assertEqual(mynodemanager.get_num_left_urls(),4)
        self.assertEqual(mynodemanager.get_num_left_trusted(),4)
        self.assertEqual(mynodemanager.get_num_left_open(),0)
        self.assertTrue(mynodemanager.has_more_urls())

    '''
    Test whether setting and resetting nodes works.
    '''
    @unittest.skipIf(globalvar.RABBIT_OPEN_NOT_ALLOWED, '(this test uses open rabbit nodes)')
    def test_setting_and_resetting_nodes(self):

        # Make a node manager with several nodes, and remove them
        # one by one.
        mynodemanager = esgfpid.rabbit.nodemanager.NodeManager()
        mynodemanager.add_trusted_node(**TESTHELPERS.get_args_for_nodemanager())
        mynodemanager.add_trusted_node(**TESTHELPERS.get_args_for_nodemanager())
        mynodemanager.add_trusted_node(**TESTHELPERS.get_args_for_nodemanager())
        mynodemanager.add_open_node(**TESTHELPERS.get_args_for_nodemanager())
        mynodemanager.add_open_node(**TESTHELPERS.get_args_for_nodemanager())

        # Run code to be tested:
        # First, set one after the other, and check if the
        # number of URLs left is correct.
        # When all are used, reset the hosts and check if
        # the resetting worked fine.

        # We added five hosts, 3 trusted and 2 untrusted:
        self.assertEqual(mynodemanager.get_num_left_urls(),5)
        self.assertEqual(mynodemanager.get_num_left_trusted(),3)
        self.assertEqual(mynodemanager.get_num_left_open(),2)
        self.assertTrue(mynodemanager.has_more_urls())
        # Set the first host (1/5), 4 are left. (First, the trusted ones are selected):
        mynodemanager.set_next_host()
        self.assertEqual(mynodemanager.get_num_left_urls(),4)
        self.assertEqual(mynodemanager.get_num_left_trusted(),2)
        self.assertEqual(mynodemanager.get_num_left_open(),2)
        self.assertTrue(mynodemanager.has_more_urls())
        # Set the second host (2/5), 3 are left.
        mynodemanager.set_next_host()
        self.assertEqual(mynodemanager.get_num_left_urls(),3)
        self.assertEqual(mynodemanager.get_num_left_trusted(),1)
        self.assertEqual(mynodemanager.get_num_left_open(),2)
        # Set the third host (3/5), 2 are left (the two untrusted ones):
        mynodemanager.set_next_host()
        self.assertEqual(mynodemanager.get_num_left_urls(),2)
        self.assertEqual(mynodemanager.get_num_left_trusted(),0)
        self.assertEqual(mynodemanager.get_num_left_open(),2)
        self.assertTrue(mynodemanager.has_more_urls())
        # Set the fourth host (4/5), 1 is left:
        mynodemanager.set_next_host()
        self.assertEqual(mynodemanager.get_num_left_urls(),1)
        self.assertEqual(mynodemanager.get_num_left_trusted(),0)
        self.assertEqual(mynodemanager.get_num_left_open(),1)
        self.assertTrue(mynodemanager.has_more_urls())
        # Set the last host (5/5), so 0 are left:
        mynodemanager.set_next_host()
        self.assertEqual(mynodemanager.get_num_left_urls(),0)
        self.assertEqual(mynodemanager.get_num_left_trusted(),0)
        self.assertEqual(mynodemanager.get_num_left_open(),0)
        self.assertFalse(mynodemanager.has_more_urls())
        # Reset all nodes.
        # This includes the first call to set_next_host(), so it 
        # sets the first of five, so 4 are left:
        mynodemanager.reset_nodes()
        self.assertEqual(mynodemanager.get_num_left_urls(),4)
        self.assertEqual(mynodemanager.get_num_left_trusted(),2)
        self.assertEqual(mynodemanager.get_num_left_open(),2)
        self.assertTrue(mynodemanager.has_more_urls())

    '''
    Test getter for the routing key suffix (1/3).
    '''
    @unittest.skipIf(globalvar.RABBIT_OPEN_NOT_ALLOWED, '(this test uses open rabbit nodes)')
    def test_adapting_routing_key_open_only(self):

        # Test variables:
        # Make a node manager with only open nodes
        mynodemanager = esgfpid.rabbit.nodemanager.NodeManager()
        mynodemanager.add_open_node(**self.__get_args_dict())
        mynodemanager.add_open_node(**self.__get_args_dict())
        # Pre-check:
        self.assertEqual(mynodemanager.get_num_left_trusted(),0, 'Failing precheck.')
        self.assertEqual(mynodemanager.get_num_left_open(),2, 'Failing precheck.')
        # A node has to be set currently for this to work:
        mynodemanager.set_next_host()
        #print(str(mynodemanager._NodeManager__current_node))

        # Run code to be tested:
        routing_key = 'foo.fresh.foo'
        routing_key = mynodemanager.adapt_routing_key_for_untrusted(routing_key)

        # Check result:
        # Check if adapt_routing_key_for_untrusted() returns
        # the correct routing key. If there is only open
        # nodes, it must indicate that it's open nodes only.
        self.assertEqual('foo.fresh-untrusted-only.foo', routing_key)


    '''
    Test getter for the routing key suffix (2/3).
    '''
    def test_adapting_routing_key_trusted_only(self):

        # Test variables:
        # Make a node manager with only open nodes
        mynodemanager = esgfpid.rabbit.nodemanager.NodeManager()
        mynodemanager.add_trusted_node(**TESTHELPERS.get_args_for_nodemanager())
        mynodemanager.add_trusted_node(**TESTHELPERS.get_args_for_nodemanager())
        # Pre-check:
        self.assertEqual(mynodemanager.get_num_left_trusted(),2, 'Failing precheck.')
        self.assertEqual(mynodemanager.get_num_left_open(),0, 'Failing precheck.')
        # A node has to be set currently for this to work:
        mynodemanager.set_next_host()
        #print(str(mynodemanager._NodeManager__current_node))

        # Run code to be tested
        routing_key = 'foo.fresh.foo'
        routing_key = mynodemanager.adapt_routing_key_for_untrusted(routing_key)

        # Check result:
        # Check if adapt_routing_key_for_untrusted() returns
        # the correct routing key.
        # If there is trusted nodes, and a trusted node is
        # currently set, it must not change the routing key.
        self.assertEqual('foo.fresh.foo', routing_key)


    '''
    Test getter for parameters for current host.
    '''
    @unittest.skipIf(globalvar.RABBIT_OPEN_NOT_ALLOWED, '(this test uses open rabbit nodes)')
    def test_get_params(self):

        # Test variables:
        mynodemanager = esgfpid.rabbit.nodemanager.NodeManager()
        mynodemanager.add_trusted_node(**TESTHELPERS.get_args_for_nodemanager(host='foo'))
        mynodemanager.add_open_node(**TESTHELPERS.get_args_for_nodemanager(host='bar'))

        # Run code to be tested (1/2):
        #mynodemanager.set_next_host() # This is called by the getter!!
        params = mynodemanager.get_connection_parameters()

        # Check result (1/2):
        self.assertIsInstance(params, pika.ConnectionParameters)
        self.assertEqual(params.host, 'foo')

        # Run code to be tested (2/2):
        mynodemanager.set_next_host()
        params = mynodemanager.get_connection_parameters()

        # Check result (2/2):
        self.assertIsInstance(params, pika.ConnectionParameters)
        self.assertEqual(params.host, 'bar')


    '''
    Test simple getter.
    '''
    def test_get_props(self):

        # Test variables:
        mynodemanager = esgfpid.rabbit.nodemanager.NodeManager()

        # Run code to be tested:
        props = mynodemanager.get_properties_for_message_publications()

        # Check result
        self.assertIsInstance(props, pika.BasicProperties)



    '''
    Test setting the priority to the lowest value, once.
    '''
    def test_change_prio(self):

        # Test variables:
        mynodemanager = esgfpid.rabbit.nodemanager.NodeManager()
        mynodemanager.add_trusted_node(**TESTHELPERS.get_args_for_nodemanager())
        mynodemanager.set_next_host()

        # Run code to be tested:
        old_prio = mynodemanager._get_prio_where_current_is_stored()
        mynodemanager.set_priority_low_for_current()
        new_prio = mynodemanager._get_prio_where_current_is_stored()

        # Check result
        self.assertEqual(old_prio, 'zzz_default', 'Default prio is %s, expected zzz_default' % old_prio)
        self.assertEqual(new_prio, 'zzzz_last',   'Prio after changing prio is %s, expected zzzz_last' % new_prio)



    '''
    Test setting the priority to the lowest value, once.
    Check if the password is masked from log output.
    '''
    def test_change_prio_pw_masking(self):

        # Test variables:
        mynodemanager = esgfpid.rabbit.nodemanager.NodeManager()
        mynodemanager.add_trusted_node(**TESTHELPERS.get_args_for_nodemanager())
        mynodemanager.set_next_host()

        # Run code to be tested:
        with self.assertLogs() as context_manager:
            mynodemanager.set_priority_low_for_current()

        # Check result
        self.assertTrue('pw_foo' not in str(context_manager.output), 'Password (pw_foo) is not masked: %s' % context_manager.output)
        self.assertTrue('pw_' in str(context_manager.output), 'Masked password stub (pw_f) missing: %s' % context_manager.output)


    '''
    Test setting the priority to the lowest value, twice.
    This tests whether the function can handle nodes that "already"
    have been set to the lowest prio.
    '''
    def test_change_prio_twice(self):

        # Test variables:
        mynodemanager = esgfpid.rabbit.nodemanager.NodeManager()
        mynodemanager.add_trusted_node(**TESTHELPERS.get_args_for_nodemanager())
        mynodemanager.set_next_host()

        # Run code to be tested:
        old_prio = mynodemanager._get_prio_where_current_is_stored()
        mynodemanager.set_priority_low_for_current()
        new_prio1 = mynodemanager._get_prio_where_current_is_stored()
        mynodemanager.set_priority_low_for_current()
        new_prio2 = mynodemanager._get_prio_where_current_is_stored()

        # Check result
        new_prio = mynodemanager._NodeManager__current_node['priority']
        self.assertEqual(old_prio, 'zzz_default', 'Default prio is %s, expected zzz_default' % old_prio)
        self.assertEqual(new_prio1, 'zzzz_last',   'Prio after changing prio is %s, expected zzzz_last' % new_prio1)
        self.assertEqual(new_prio2, 'zzzz_last',   'Prio after changing prio is %s, expected zzzz_last' % new_prio2)

    '''
    Test setting the priority to a weird nonexisting value
    This is only to test the code that should catch this - I doubt
    that this could ever ever every happen.
    '''
    def test_change_prio_weird_prio(self):

        # Test variables:
        mynodemanager = esgfpid.rabbit.nodemanager.NodeManager()
        mynodemanager.add_trusted_node(**TESTHELPERS.get_args_for_nodemanager())
        mynodemanager.set_next_host()
        # Change the prio to a weird value, to check if the functions still work:
        mynodemanager._NodeManager__trusted_nodes_archive['dummy_prio'] = [mynodemanager._NodeManager__current_node]
        del mynodemanager._NodeManager__trusted_nodes_archive['zzz_default']

        # Run code to be tested:
        old_prio = mynodemanager._get_prio_where_current_is_stored()
        mynodemanager.set_priority_low_for_current()
        new_prio1 = mynodemanager._get_prio_where_current_is_stored()
        mynodemanager.set_priority_low_for_current()
        new_prio2 = mynodemanager._get_prio_where_current_is_stored()

        # Check result
        new_prio = mynodemanager._NodeManager__current_node['priority']
        self.assertEqual(old_prio, 'dummy_prio', 'Default prio is %s, expected dummy_prio' % old_prio)
        self.assertEqual(new_prio1, 'zzzz_last',  'Prio after changing prio is %s, expected zzzz_last' % new_prio1)
        self.assertEqual(new_prio2, 'zzzz_last',  'Prio after changing prio is %s, expected zzzz_last' % new_prio2)




import logging
import esgfpid.utils
import esgfpid.assistant.messages
from esgfpid.utils import loginfo, logdebug, logtrace, logerror, logwarn

LOGGER = logging.getLogger(__name__)
LOGGER.addHandler(logging.NullHandler())

'''
This class is responsible for creating data cart handles.

Its API consists of one method, make_data_cart_pid().

Dependencies:
 - esgfpid.coupling.Coupler object, to send the JSON message to RabbitMQ.
 - esgfpid.assistant.messages module to create the JSON message.
 - esgfpid.utils module to create the handle string.
'''
class DataCartAssistant(object):

    '''
    Constructor for a data cart assistant.

    :param prefix:  The handle prefix, used for creating a handle
                    string for this data cart.
    :param coupler: The coupler object, used to send the message
                    to RabbitMQ.
    '''
    def __init__(self, **args):
        mandatory_args = ['prefix', 'coupler']
        esgfpid.utils.check_presence_of_mandatory_args(args, mandatory_args)
        esgfpid.utils.check_noneness_of_mandatory_args(args, mandatory_args)
        self.__prefix = args['prefix']
        self.__coupler = args['coupler']

    '''
    Create a handle string for a data cart, based
    on its content, so that with the same content,
    the handle string is always the same.

    :param dict_of_drs_ids_and_pids:
    :param prefix:
    '''
    @staticmethod
    def _get_handle_string_for_datacart(dict_of_drs_ids_and_pids, prefix):
        list_of_drs_ids = list(dict_of_drs_ids_and_pids.keys())
        hash_basis = esgfpid.utils.make_sorted_lowercase_list_without_hdl(list_of_drs_ids)
        return esgfpid.utils.make_handle_from_list_of_strings(hash_basis, prefix, addition='datacart')
        # This sorts the list, removes all "hdl:", and makes a hash
    
    '''
    Create a request to make a handle for a data cart, and send 
    it to the RabbitMQ.

    :param dict_of_drs_ids_and_pids: A dictionary of all the 
            dataset ids and dataset pids that should be contained
            in this 
    '''
    def make_data_cart_pid(self, dict_of_drs_ids_and_pids):
        logdebug(LOGGER, 'Making a PID for a data cart full of datasets...')

        # Check arg
        if not type(dict_of_drs_ids_and_pids) == type(dict()):
            if type(dict_of_drs_ids_and_pids) == type([]):
                raise esgfpid.exceptions.ArgumentError('Please provide a dictionary of dataset ids and handles, not a list')
            else:
                raise esgfpid.exceptions.ArgumentError('Please provide a dictionary of dataset ids and handles')

        # Make a pid (hash on the content):
        cart_handle = DataCartAssistant._get_handle_string_for_datacart(
            dict_of_drs_ids_and_pids,
            self.__prefix
        )

        # Make and send message
        message = self.__make_message(cart_handle, dict_of_drs_ids_and_pids)
        self.__send_message_to_queue(message)

        # Return pid
        logdebug(LOGGER, 'Making a PID for a data cart full of datasets... done.')
        loginfo(LOGGER, 'Requesting to create PID for data cart (%s).', cart_handle)
        return cart_handle

    '''
    Create the JSON message for creating the data cart PID.
    This calls the esgfpid.assistant.messages module.
    '''
    def __make_message(self, cart_handle, dict_of_drs_ids_and_pids):
        message_timestamp = esgfpid.utils.get_now_utc_as_formatted_string()
        message = esgfpid.assistant.messages.make_data_cart_message(
            cart_handle = cart_handle,
            timestamp = message_timestamp,
            data_cart_content = dict_of_drs_ids_and_pids
        )
        return message

    '''
    Send the JSON message to the RabbitMQ, using the
    esgfpid.coupling module.
    '''
    def __send_message_to_queue(self, message):
        self.__coupler.send_message_to_queue(message)


import unittest
import mock
import logging
import json
import esgfpid.solr.solr
import esgfpid.solr.tasks.all_versions_of_dataset as task
import tests.resources.responsemock
import tests.utils

# Logging:
LOGGER = logging.getLogger(__name__)
LOGGER.addHandler(logging.NullHandler())

# Test resources:
from resources.TESTVALUES import *
import resources.TESTVALUES as TESTHELPERS

# Load some data that is needed for testing
PATH_RES = tests.utils.get_super_neighbour_directory(__file__, 'resources')
SOLR_RESPONSE = json.load(open(PATH_RES+'/solr_response.json'))


QUERY = {'format': 'application/solr+json', 'facets': 'pid,version', 'limit': 0, 'distrib': False, 'drs_id':'abc', 'type': 'Dataset'}


class SolrTask2TestCase(unittest.TestCase):

    def setUp(self):
        LOGGER.info('######## Next test (%s) ##########', __name__)

    def tearDown(self):
        LOGGER.info('#############################')

    def make_testtask(self):
        testsolr = TESTHELPERS.get_testsolr()
        testtask = task.FindVersionsOfSameDataset(testsolr)
        return testtask

    def get_args_dict(self):
        return dict(
            drs_id = 'abc',
            version_number = '2016',
            data_node = 'foo.de',
            prefix = '123')

    def fake_solr_response(self, pid, vers):
        resp = {
            "facet_counts": {
                "facet_fields": {
                    "pid": pid,
                    "version": vers
                }
            }
        }
        return resp

    # Actual tests:

    def test_init_ok(self):

        # Preparations
        testsolr = TESTHELPERS.get_testsolr()

        # Run code to be tested:
        testtask = task.FindVersionsOfSameDataset(testsolr)

        # Check result
        self.assertIsInstance(testtask, task.FindVersionsOfSameDataset, 'Constructor fail.')

    @mock.patch('esgfpid.solr.serverconnector.SolrServerConnector.send_query')
    def test_retrieve_dataset_handles_both_returned_ok(self, getpatch):

        # Define the replacement for the patched method:
        pids = ["123/456",3,"123/234",1,"987/567",2]
        versions = ["2016",3,"2015",7,"2099",8]
        getpatch.return_value = self.fake_solr_response(pids, versions)

        # Preparations
        task = self.make_testtask()

        # Test variables:
        drs_id = 'abc'
        prefix = '123'

        # Run code to be tested:
        received_dict = task.retrieve_dataset_handles_or_version_numbers_of_all_versions(drs_id, prefix)

        # Check result:
        # Was the correct query sent?
        expected_query = QUERY
        getpatch.assert_called_once_with(expected_query)
        # Was the response treated correctly?
        expected_dict = {
            'version_numbers': ['2015', '2016', '2099'],
            'dataset_handles': ['hdl:123/987/567', 'hdl:123/234', 'hdl:123/456']
        }
        self.assertEqual(sorted(expected_dict), sorted(received_dict), 'Expected %s, but got %s' % (expected_dict, received_dict))

    @mock.patch('esgfpid.solr.serverconnector.SolrServerConnector.send_query')
    def test_retrieve_dataset_handles_only_handles_returned_ok(self, getpatch):

        # Define the replacement for the patched method:
        pids = ["123/456",3,"123/234",1,"987/567",2]
        versions = []
        getpatch.return_value = self.fake_solr_response(pids, versions)

        # Preparations
        task = self.make_testtask()

        # Test variables:
        drs_id = 'abc'
        prefix = '123'

        # Run code to be tested:
        received_dict = task.retrieve_dataset_handles_or_version_numbers_of_all_versions(drs_id, prefix)

        # Check result:
        # Was the correct query sent?
        expected_query = QUERY
        getpatch.assert_called_once_with(expected_query)
        # Was the response treated correctly?
        expected_dict = {
            'version_numbers': [],
            'dataset_handles': ['hdl:123/987/567', 'hdl:123/234', 'hdl:123/456']
        }
        self.assertEqual(sorted(expected_dict), sorted(received_dict), 'Expected %s, but got %s' % (expected_dict, received_dict))



    @mock.patch('esgfpid.solr.serverconnector.SolrServerConnector.send_query')
    def test_retrieve_dataset_handles_only_versions_returned_ok(self, getpatch):

        # Define the replacement for the patched method:
        pids = []
        versions = ["2016",3,"2015",7,"2099",8]
        getpatch.return_value = self.fake_solr_response(pids, versions)

        # Preparations
        task = self.make_testtask()

        # Test variables:
        drs_id = 'abc'
        prefix = '123'

        # Run code to be tested:
        received_dict = task.retrieve_dataset_handles_or_version_numbers_of_all_versions(drs_id, prefix)

        # Check result:
        # Was the correct query sent?
        expected_query = QUERY
        getpatch.assert_called_once_with(expected_query)
        # Was the response treated correctly?
        expected_dict = {
            'version_numbers': ['2015', '2016', '2099'],
            'dataset_handles': []
        }
        self.assertEqual(sorted(expected_dict), sorted(received_dict), 'Expected %s, but got %s' % (expected_dict, received_dict))



    @mock.patch('esgfpid.solr.serverconnector.SolrServerConnector.send_query')
    def test_retrieve_dataset_handles_emtpy_lists_returned_ok(self, getpatch):

        # Define the replacement for the patched method:
        pids = []
        versions = []
        getpatch.return_value = self.fake_solr_response(pids, versions)

        # Preparations
        task = self.make_testtask()

        # Test variables:
        drs_id = 'abc'
        prefix = '123'

        # Run code to be tested:
        received_dict = task.retrieve_dataset_handles_or_version_numbers_of_all_versions(drs_id, prefix)

        # Check result:
        # Was the correct query sent?
        expected_query = QUERY
        getpatch.assert_called_once_with(expected_query)
        # Was the response treated correctly?
        expected_dict = {
            'version_numbers': [],
            'dataset_handles': []
        }
        self.assertEqual(sorted(expected_dict), sorted(received_dict), 'Expected %s, but got %s' % (expected_dict, received_dict))


    @mock.patch('esgfpid.solr.serverconnector.SolrServerConnector.send_query')
    def test_retrieve_dataset_handles_solr_error(self, getpatch):

        # Define the replacement for the patched method:
        getpatch.side_effect = esgfpid.exceptions.SolrError()

        # Preparations
        task = self.make_testtask()

        # Test variables:
        drs_id = 'abc'
        prefix = '123'

        # Run code to be tested:
        with self.assertRaises(esgfpid.exceptions.SolrError):
            received_dict = task.retrieve_dataset_handles_or_version_numbers_of_all_versions(drs_id, prefix)






import logging
import pika
import esgfpid.utils
from esgfpid.utils import logwarn
from .nodemanager import NodeManager
from .asynchronous import AsynchronousRabbitConnector
from .synchronous import SynchronousRabbitConnector


# Normal logger:
LOGGER = logging.getLogger(__name__)
LOGGER.addHandler(logging.NullHandler())

'''
This class provides the library with a facade to
a module that handles connections to RabbitMQ in
a synchronous or asynchronous way.

It is basically a facade/wrapper for either a
:py:class:`~esgfpid.rabbit.synchronous.SynchronousServerConnector`
or a 
:py:class:`~esgfpid.rabbit.asynchronous.asynchronous.AsynchronousServerConnector`

'''
class RabbitMessageSender(object):

    '''
    Create a RabbitMessageSender that takes care of managing
    connections to RabbitMQ instances and sending messages 
    to them.

    :param exchange_name: Mandatory. The name of the exchange
        to send messages to (string).
    :param credentials: Mandatory. List of dictionaries containing
        the information about the RabbitMQ nodes.
    :param is_synchronous_mode: Mandatory. Boolean to define if
        the connection to RabbitMQ and the message sending
        should work in synchronous mode.
    :param test_publication: Mandatory. Boolean to tell whether
        a test flag should be added to all messages.

    '''
    def __init__(self, **args):
        LOGGER.debug('Initializing RabbitMessageSender.')

        mandatory_args = [
            'exchange_name',
            'credentials',
            'test_publication',
            'is_synchronous_mode'
        ]
        esgfpid.utils.check_presence_of_mandatory_args(args, mandatory_args)

        self.__ASYNCHRONOUS = not args['is_synchronous_mode']
        self.__test_publication = args['test_publication']
        self.__node_manager = self.__make_rabbit_settings(args)
        self.__server_connector = self.__init_server_connector(args, self.__node_manager)

    def __init_server_connector(self, args, node_manager):
        if self.__ASYNCHRONOUS:
            return esgfpid.rabbit.asynchronous.AsynchronousRabbitConnector(node_manager)
        else:
            return esgfpid.rabbit.synchronous.SynchronousRabbitConnector(node_manager)


    '''
    Open a synchronous connection to RabbitMQ.

    This only has an effect if the RabbitMessageSender is in
    synchronous mode.

    Please see: :func:`~rabbit.synchronous.SynchronousServerConnector.open_rabbit_connection`.

    If this is not called before the first message is sent,
    the first sent message automatically opens a connection.

    Opening and closing the connection is provided as
    separate methods because clients that know that they
    want to send many messages one after the other should
    preferably open and close the synchronous connection
    themselves, to avoid opening and reclosing all the time.

    This is called for example by the publish assistant,
    which opens a connection, sends all the messages that
    belong to one dataset, and closes it again.
    '''
    def open_rabbit_connection(self):
        if not self.__ASYNCHRONOUS:
            return self.__server_connector.open_rabbit_connection()

    '''
    Close a synchronous connection to RabbitMQ.

    This only has an effect if the RabbitMessageSender is in
    synchronous mode.

    Synchronous connections are never closed automatically.
    However, if the client forgets to close, it does not
    block the library. It is still preferred that the client
    closes the connection as soon as he knows he won't send
    any more messages (for a while).
    Otherwise, the connection is kept open and will send
    heartbeats from time to time.

    This is called for example by the publish assistant,
    which opens a connection, sends all the messages that
    belong to one dataset, and closes it again.
    '''
    def close_rabbit_connection(self):
        if not self.__ASYNCHRONOUS:
            return self.__server_connector.close_rabbit_connection()


    '''

    Close a thread that communicates with RabbitMQ asynchronously,
    in a gentle way. This means that the thread does try to send
    the last pending messages and receive the last pending
    delivery confirmations, up to a maximum waiting time.

    This only has an effect if the RabbitMessageSender is in
    asynchronous mode.

    Please see documentation of asynchronous rabbit module
    (:func:`~rabbit.asynchronous.asynchronous.AsynchronousRabbitConnector.force_finish_rabbit_thread`).
    '''
    def finish(self):
        if self.__ASYNCHRONOUS:
            self.__server_connector.finish_rabbit_thread()

    def is_finished(self):
        if self.__ASYNCHRONOUS:
            return self.__server_connector.is_finished()
        else:
            return None

    '''
    Start a thread that communicates with RabbitMQ asynchronously.

    This only has an effect if the RabbitMessageSender is in
    asynchronous mode.

    Please see documentation of asynchronous rabbit module
    (:func:`~rabbit.asynchronous.asynchronous.AsynchronousRabbitConnector.start_rabbit_thread`).
    '''
    def start(self):
        if self.__ASYNCHRONOUS:
            self.__server_connector.start_rabbit_thread()

    '''
    Force-close a thread that communicates with RabbitMQ asynchronously.

    This only has an effect if the RabbitMessageSender is in
    asynchronous mode.
    Please see documentation of asynchronous rabbit module
    (:func:`~rabbit.asynchronous.asynchronous.AsynchronousRabbitConnector.force_finish_rabbit_thread`).
    '''
    def force_finish(self):
        if self.__ASYNCHRONOUS:
            self.__server_connector.force_finish_rabbit_thread()

    def any_leftovers(self):
        if self.__ASYNCHRONOUS:
            return self.__server_connector.any_leftovers()

    def get_leftovers(self):
        if self.__ASYNCHRONOUS:
            return self.__server_connector.get_leftovers()

    '''
    Send a message to RabbitMQ.

    In asynchronous mode, we cannot tell whether the
    delivery as successful, as the delivery confirmation
    will arrive later.

    In synchronous mode, if the delivery was not successful,
    an exception is raised.

    :param: JSON message as string or dictionary. It should
        include its routing key as a dictionary entry with
        key "ROUTING_KEY", Otherwise a default routing key
        will be used to send the message.
    :raises: esgfpid.exceptions.MessageNotDeliveredException:
        In case the message was not delivered. Only in
        synchronous mode.
    '''
    def send_message_to_queue(self, message):
        if self.__test_publication == True:
            message['test_publication'] = True
        self.__server_connector.send_message_to_queue(message)

    def __make_rabbit_settings(self, args):
        node_manager = NodeManager()

        # Add all RabbitMQ nodes:
        for cred in args['credentials']:
            
            if 'priority' not in cred:
                cred['priority'] = None
            if 'vhost' not in cred:
                cred['vhost'] = None
            if 'port' not in cred:
                cred['port'] = None
            if 'ssl_enabled' not in cred:
                cred['ssl_enabled'] = None

            # Open node:
            if cred['password'] == 'jzlnL78ZpExV#_QHz':
                node_manager.add_open_node(
                    username=cred['user'],
                    password='U6-Lke39mN',
                    host=cred['url'],
                    exchange_name=args['exchange_name'],
                    priority=cred['priority'],
                    vhost=cred['vhost'],
                    port=cred['port'],
                    ssl_enabled=cred['ssl_enabled']
                )

            # Trusted node:
            else:
                node_manager.add_trusted_node(
                    username=cred['user'],
                    password=cred['password'],
                    host=cred['url'],
                    exchange_name=args['exchange_name'],
                    priority=cred['priority'],
                    vhost=cred['vhost'],
                    port=cred['port'],
                    ssl_enabled=cred['ssl_enabled']
                )

        return node_manager



class MockRequest(object):
    '''
    This is a mocked Request object containing only an url,
    as this is the only attribute accessed during the tests.
    There is a default value for it, but it can also be passed.
    '''
    def __init__(self, url=None):
        if url is not None:
            self.url = url
        else:
            self.url = 'http://foo.foo'

class MockSolrResponse(object):
    '''
    This is a mocked Response object (can be used to replace
    a response from any call to "requests.get" or
    "request.put" or "request.delete", ...).

    It contains a request, a status code and some JSON content.
    For all of these, there is default values, but they can also
    be passed.

    Some standard cases are available, e.g. or "handle not found",
    which has a specific combination of HTTP status code, handle
    response code and content.
    '''
    def __init__(self, status_code=None, content=None, request=None, success=False, notfound=False, empty=False):

        self.content = None
        self.status_code = None
        self.request = None

        # Some predefined cases:
        if success:
            self.status_code = 200
            self.content = '{"responseHeader":{}, "response":{}, "facet_counts": {"facet_fields": {"bla": ["blub",1,"miau",4]}}}'
        elif notfound:
            self.status_code = 404
            self.content = ''
        # User-defined overrides predefined cases:
        if content is not None:
            self.content = content
        if status_code is not None:
            self.status_code = status_code
        if request is not None:
            self.request = request
        # Defaults (they do not override):
        if self.content is None:    
            self.content = '{"responseHeader":{}, "response":{}, "facet_counts": {}}'
        if self.status_code is None:
            self.status_code = 200
        if self.request is None:
            self.request = MockRequest()
        # Special case: Content should be None:
        if self.content is 'NONE':
            self.content = None



import unittest
import mock
import logging
import json
import requests
import esgfpid.solr.solr
import tests.resources.responsemock
import tests.resources

# Logging:
LOGGER = logging.getLogger(__name__)
LOGGER.addHandler(logging.NullHandler())

# Load some data that is needed for testing
PATH_RES = tests.utils.get_super_neighbour_directory(__file__, 'resources')
SOLR_RESPONSE = json.load(open(PATH_RES+'/solr_response.json'))

# Test resources:
from resources.TESTVALUES import *
import resources.TESTVALUES as TESTHELPERS

class SolrTestCase(unittest.TestCase):

    def setUp(self):
        LOGGER.info('######## Next test (%s) ##########', __name__)

    def tearDown(self):
        LOGGER.info('#############################')

    def test_init_switched_off_ok(self):

        # Run code to be tested:
        testsolr = esgfpid.solr.SolrInteractor(
            solr_url = SOLR_URL_LIBRARY,
            prefix = PREFIX_NO_HDL,
            switched_off = True,
            https_verify = True,
            disable_insecure_request_warning = False
        )
        has_access = not testsolr.is_switched_off()

        # Check result:
        self.assertIsInstance(testsolr, esgfpid.solr.SolrInteractor, 'Constructor fail.')
        self.assertFalse(has_access, 'Solr pretends to have access although we told it to switch off.')

    def test_init_ok(self):

        # Run code to be tested:
        testsolr = esgfpid.solr.SolrInteractor(
            solr_url = SOLR_URL_LIBRARY,
            prefix = PREFIX_NO_HDL,
            switched_off = False,
            https_verify = True,
            disable_insecure_request_warning = False
        )
        has_access = not testsolr.is_switched_off()

        # Check result:
        self.assertIsInstance(testsolr, esgfpid.solr.SolrInteractor, 'Constructor fail.')
        self.assertTrue(has_access, 'Solr pretends not to have access.')

    def test_init_missing_argument(self):

        # Run code to be tested and check exception:
        with self.assertRaises(esgfpid.exceptions.ArgumentError):
            testsolr = esgfpid.solr.SolrInteractor(
                solr_url = None,
                prefix = PREFIX_NO_HDL
            )

    def test_make_solr_base_query_ok(self):

        # Run code to be tested:
        testsolr = TESTHELPERS.get_testsolr_switched_off()
        query = testsolr.make_solr_base_query()

        # Check result:
        expected = dict(
            distrib=False,
            format='application/solr+json',
            limit=0
        )
        self.assertTrue(query==expected,
            'Make solr base query: Unexpected query made!\nExpected: '+str(expected)+'\nReceived: '+str(query)+'.')

    def test_send_query_switched_off(self):

        # Test variables
        testsolr = TESTHELPERS.get_testsolr_switched_off()
        query = testsolr.make_solr_base_query()

        # Run code to be tested:
        with self.assertRaises(esgfpid.exceptions.SolrSwitchedOff):
            response = testsolr.send_query(query)

    @mock.patch('esgfpid.solr.serverconnector.requests.get')
    def test_send_query_ok_patched(self, getpatch):

        # Define the replacement for the patched method:
        mock_response = tests.resources.responsemock.MockSolrResponse(success=True)
        getpatch.return_value = mock_response

        # Test variables
        testsolr = TESTHELPERS.get_testsolr()
        query = testsolr.make_solr_base_query()

        # Run code to be tested:
        response = testsolr.send_query(query)

        # Check result:
        # We expect the mock response to be returned and parsed to JSON.
        self.assertTrue('responseHeader' in response, 'Solr response was not correct JSON: '+str(response))
        self.assertTrue(len(response['facet_counts']['facet_fields']['bla'])==4, 'JSON content was not transmitted correctly.')

    @mock.patch('esgfpid.solr.serverconnector.requests.get')
    def test_send_query_404_patched(self, getpatch):

        # Define the replacement for the patched method:
        mock_response = tests.resources.responsemock.MockSolrResponse(notfound=True)
        getpatch.return_value = mock_response

        # Test variables
        testsolr = TESTHELPERS.get_testsolr()
        query = testsolr.make_solr_base_query()

        # Run code to be tested and check exception:
        with self.assertRaises(esgfpid.exceptions.SolrError) as raised:
            response = testsolr.send_query(query)
        self.assertIn('HTTP 404', repr(raised.exception))


    @mock.patch('esgfpid.solr.serverconnector.requests.get')
    def test_send_query_5000_patched(self, getpatch):

        # Define the replacement for the patched method:
        mock_response = tests.resources.responsemock.MockSolrResponse(status_code=5000)
        getpatch.return_value = mock_response

        # Test variables
        testsolr = TESTHELPERS.get_testsolr()
        query = testsolr.make_solr_base_query()

        # Run code to be tested and check exception:
        with self.assertRaises(esgfpid.exceptions.SolrError) as raised:
            response = testsolr.send_query(query)
        self.assertIn('code 5000', repr(raised.exception))


    def test_retrieve_file_handles_of_same_dataset_no_access(self):

        # Test variables
        testsolr = TESTHELPERS.get_testsolr(switched_off=True)

        # Run code to be tested and check exception:
        with self.assertRaises(esgfpid.exceptions.SolrSwitchedOff):
            resp = testsolr.retrieve_file_handles_of_same_dataset(
                drs_id=DRS_ID,
                version_number=DS_VERSION,
                data_node=DATA_NODE
            )


    @mock.patch('esgfpid.solr.serverconnector.requests.get')
    def test_retrieve_file_handles_of_same_dataset_patched(self, getpatch):

        # Define the replacement for the patched method:
        prefix = PREFIX_NO_HDL
        response_json = {
            "responseHeader":{},
            "response":{},
            "facet_counts": {
                "facet_fields":{
                    "tracking_id": [
                        prefix+'/123',1,
                        prefix+'/234',1,
                        prefix+'/345',1
                    ]
                }
            }
        }
        mock_response = tests.resources.responsemock.MockSolrResponse(success=True, content=json.dumps(response_json))
        getpatch.return_value = mock_response

        # Test variables
        testsolr = TESTHELPERS.get_testsolr()

        # Run code to be tested:
        resp = testsolr.retrieve_file_handles_of_same_dataset(
            drs_id=DRS_ID,
            version_number=DS_VERSION,
            data_node=DATA_NODE
        )

        # Check result:
        expected = ['hdl:'+prefix+'/345', 'hdl:'+prefix+'/234', 'hdl:'+prefix+'/123']
        ok = (set(resp) == set(expected))
        self.assertTrue(ok, 'Solr returned:\n'+str(resp)+'\nExpected:\n'+str(expected))

    @mock.patch('esgfpid.solr.serverconnector.requests.get')
    def test_send_query_patched_connection_error(self, getpatch):

        # Define the replacement for the patched method:
        getpatch.side_effect = requests.exceptions.ConnectionError()

        # Test variables
        testsolr = TESTHELPERS.get_testsolr()
        query = testsolr.make_solr_base_query()

        # Run code to be tested and check exception:
        with self.assertRaises(esgfpid.exceptions.SolrError) as raised:
            response = testsolr.send_query(query)
        self.assertIn('ConnectionError', repr(raised.exception))

    @mock.patch('esgfpid.solr.serverconnector.requests.get')
    def test_retrieve_datasethandles_or_versionnumbers_of_allversions_pids_patched(self, getpatch):

        # Define the replacement for the patched method:
        prefix = PREFIX_NO_HDL
        response_json = {
            "responseHeader":{},
            "response":{},
            "facet_counts": {
                "facet_fields":{
                    "pid": [
                        prefix+'/123',1,
                        prefix+'/234',1,
                        prefix+'/345',1
                    ]
                }
            }
        }
        mock_response = tests.resources.responsemock.MockSolrResponse(success=True, content=json.dumps(response_json))
        getpatch.return_value = mock_response

        # Test variables
        testsolr = TESTHELPERS.get_testsolr()

        # Run code to be tested:
        resp = testsolr.retrieve_datasethandles_or_versionnumbers_of_allversions(DRS_ID)

        # Check result:
        expected = {'version_numbers': None, 'dataset_handles': ['hdl:'+prefix+'/345', 'hdl:'+prefix+'/234', 'hdl:'+prefix+'/123']}
        ok = (set(resp) == set(expected))
        self.assertTrue(ok, 'Solr returned:\n'+str(resp)+'\nExpected:\n'+str(expected))

    @mock.patch('esgfpid.solr.serverconnector.requests.get')
    def test_retrieve_datasethandles_or_versionnumbers_of_allversions_vers_patched(self, getpatch):

        # Define the replacement for the patched method:
        prefix = PREFIX_NO_HDL
        response_json = {
            "responseHeader":{},
            "response":{},
            "facet_counts": {
                "facet_fields":{
                    "version": [
                        '201611',1,
                        '201622',1,
                        '201633',1,
                    ]
                }
            }
        }
        mock_response = tests.resources.responsemock.MockSolrResponse(success=True, content=json.dumps(response_json))
        getpatch.return_value = mock_response

        # Test variables
        testsolr = TESTHELPERS.get_testsolr()

        # Run code to be tested:
        resp = testsolr.retrieve_datasethandles_or_versionnumbers_of_allversions(DRS_ID)

        # Check result:
        expected = {'version_numbers': ['201622', '201611', '201633'], 'dataset_handles': None}

        ok = (set(resp) == set(expected))
        self.assertTrue(ok, 'Solr returned:\n'+str(resp)+'\nExpected:\n'+str(expected))


    @mock.patch('esgfpid.solr.serverconnector.requests.get')
    def test_retrieve_datasethandles_or_versionnumbers_of_allversions_both_patched(self, getpatch):

        # Define the replacement for the patched method:
        prefix = PREFIX_NO_HDL
        response_json = {
            "responseHeader":{},
            "response":{},
            "facet_counts": {
                "facet_fields":{
                    "version": [
                        '201611',1,
                        '201622',1,
                        '201633',1,
                    ],
                    "pid": [
                        prefix+'/123',1,
                        prefix+'/234',1,
                        prefix+'/345',1
                    ]
                }
            }
        }
        mock_response = tests.resources.responsemock.MockSolrResponse(success=True, content=json.dumps(response_json))
        getpatch.return_value = mock_response

        # Test variables
        testsolr = TESTHELPERS.get_testsolr()
        drs_id = DRS_ID

        # Run code to be tested:
        resp = testsolr.retrieve_datasethandles_or_versionnumbers_of_allversions(drs_id)

        # Check result:
        expected = {
            'version_numbers': ['201622', '201611', '201633'],
            'dataset_handles': ['hdl:'+prefix+'/345', 'hdl:'+prefix+'/234', 'hdl:'+prefix+'/123']
        }

        ok = (set(resp) == set(expected))
        self.assertTrue(ok, 'Solr returned:\n'+str(resp)+'\nExpected:\n'+str(expected))

    @mock.patch('esgfpid.solr.serverconnector.requests.get')
    def test_retrieve_datasethandles_or_versionnumbers_of_allversions_none_patched(self, getpatch):

        # Define the replacement for the patched method:
        prefix = PREFIX_NO_HDL
        response_json = {
            "responseHeader":{},
            "response":{},
            "facet_counts": {
                "facet_fields":{
                }
            }
        }
        mock_response = tests.resources.responsemock.MockSolrResponse(success=True, content=json.dumps(response_json))
        getpatch.return_value = mock_response

        # Test variables
        testsolr = TESTHELPERS.get_testsolr()
        drs_id = DRS_ID

        # Run code to be tested and check exception:
        with self.assertRaises(esgfpid.exceptions.SolrResponseError):
            resp = testsolr.retrieve_datasethandles_or_versionnumbers_of_allversions(drs_id)



    def test_retrieve_datasethandles_or_versionnumbers_of_allversions_no_access(self):

        # Test variables
        testsolr = TESTHELPERS.get_testsolr(switched_off=True)

        # Run code to be tested and check exception:
        with self.assertRaises(esgfpid.exceptions.SolrSwitchedOff):
            resp = testsolr.retrieve_datasethandles_or_versionnumbers_of_allversions(DRS_ID)

